# pyright: reportPrivateUsage = false

from __future__ import annotations

import ast
import builtins
import collections
import copy
import inspect
import re
import sys
import textwrap
import types
import typing
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Mapping, Optional, Sequence, Tuple, Type, TypeVar, Union, cast, overload

import pyarrow as pa

from chalk._lsp.error_builder import FeatureClassErrorBuilder, LSPErrorBuilder
from chalk.features import is_features_cls
from chalk.features._class_property import classproperty, classproperty_support
from chalk.features._encoding.pyarrow import pyarrow_to_primitive
from chalk.features.dataframe._impl import DataFrameMeta
from chalk.features.feature_field import CacheStrategy, Feature, VersionInfo
from chalk.features.feature_set import ClassSource, Features, CURRENT_FEATURE_REGISTRY
from chalk.features.feature_time import feature_time
from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
from chalk.features.namespace_context import build_namespaced_name
from chalk.features.tag import Tags
from chalk.features.underscore import Underscore
from chalk.features.underscore_features import NamedUnderscoreExpr
from chalk.serialization.parsed_annotation import ParsedAnnotation
from chalk.streams import Windowed
from chalk.streams._windows import GroupByWindowed, get_name_with_duration
from chalk.utils import notebook
from chalk.utils.collections import ensure_tuple
from chalk.utils.duration import Duration, parse_chalk_duration, parse_chalk_duration_s
from chalk.utils.metaprogramming import MISSING, set_new_attribute
from chalk.utils.string import to_snake_case

T = TypeVar("T")

GENERATED_OBSERVED_AT_NAME = "__chalk_observed_at__"

__all__ = ["features"]


@overload
def features(
    *,
    owner: Optional[str] = None,
    tags: Optional[Tags] = None,
    etl_offline_to_online: bool = False,
    max_staleness: Optional[Duration] = None,
    name: Optional[str] = None,
    singleton: bool = False,
) -> Callable[[Type[T]], Type[T]]: ...


@overload
def features(cls: Type[T]) -> Type[T]: ...


def features(
    cls: Optional[Type[T]] = None,
    *,
    owner: Optional[str] = None,
    tags: Optional[Tags] = None,
    etl_offline_to_online: bool = False,
    max_staleness: Optional[Duration] = None,
    name: Optional[str] = None,
    singleton: bool = False,
) -> Union[Callable[[Type[T]], Type[T]], Type[T]]:
    """Chalk lets you spell out your features directly in Python.

    Features are namespaced to a `FeatureSet`.
    To create a new `FeatureSet`, apply the `@features`
    decorator to a Python class with typed attributes.
    A `FeatureSet` is constructed and functions much like
    Python's own `dataclass`.

    Parameters
    ----------
    owner
        The individual or team responsible for these features.
        The Chalk Dashboard will display this field, and alerts
        can be routed to owners.
    tags
        Added metadata for features for use in filtering, aggregations,
        and visualizations. For example, you can use tags to assign
        features to a team and find all features for a given team.
    etl_offline_to_online
        When `True`, Chalk copies this feature into the online environment
        when it is computed in offline resolvers.
        Setting `etl_offline_to_online` on a feature class assigns it to all features on the
        class which do not explicitly specify `etl_offline_to_online`.
    max_staleness
        When a feature is expensive or slow to compute, you may wish to cache its value.
        Chalk uses the terminology "maximum staleness" to describe how recently a feature
        value needs to have been computed to be returned without re-running a resolver.
        Assigning a `max_staleness` to the feature class assigns it to all features on the
        class which do not explicitly specify a `max_staleness` value of their own.

    Other Parameters
    ----------------
    cls
        The decorated class. You shouldn't need to pass this argument.
    name
        The name for the feature set. By default, the name of a feature is
        taken from the name of the attribute on the class, prefixed with
        the camel-cased name of the class.
    singleton
        If `True`, the feature set is a singleton, and there will be only
        one instance of the class. Because there is only one instance, a
        singleton feature class does not have a primary key. Features
        defined on a singleton feature class are available in all resolvers,
        regardless of whatever other features are used as inputs to the resolver.

    Examples
    --------
    >>> from chalk.features import features
    >>> @features(
    ...     owner="andy@chalk.ai",
    ...     max_staleness="30m",
    ...     etl_offline_to_online=True,
    ...     tags="user-group",
    ... )
    ... class User:
    ...     id: str
    ...     # Comments here appear in the web!
    ...     # :tags: pii
    ...     name: str | None
    ...     # :owner: userteam@mycompany.com
    ...     location: LatLng
    """

    def wrap(c: Type[T]) -> Type[T]:
        namespace = name if name is not None else to_snake_case(c.__name__)
        namespace = build_namespaced_name(name=namespace)

        source_info = ClassSource(c)

        error_builder = FeatureClassErrorBuilder(
            uri=source_info.filename if source_info.filename is not None else "__main__",
            namespace=namespace,
            node=source_info and source_info.tree,
        )
        nonlocal max_staleness
        if name is not None and re.sub(r"[^a-z_0-9]", "", namespace) != namespace:
            error_builder.add_diagnostic(
                message=(
                    f"Namespace must be composed of lower-case alpha-numeric characters and '_'. Provided namespace "
                    f"'{namespace}' for class '{c.__name__}' contains invalid characters."
                ),
                code="11",
                label="invalid namespace",
                range=error_builder.decorator_kwarg_value_range(kwarg="name") or error_builder.class_definition_range(),
                raise_error=ValueError,
            )

        if name is not None and len(namespace) == 0:
            error_builder.add_diagnostic(
                message=f"Namespace cannot be an empty string, but is for the class '{c.__name__}'.",
                label="empty name",
                code="12",
                range=error_builder.decorator_kwarg_value_range(kwarg="name") or error_builder.class_definition_range(),
                raise_error=ValueError,
            )

        if max_staleness is None:
            max_staleness = timedelta(0)
        else:
            try:
                max_staleness = parse_chalk_duration(max_staleness)
            except ValueError as e:
                error_builder.add_diagnostic(
                    message=f"Invalid 'max_staleness'. {e.args[0]}",
                    label=f"invalid duration {max_staleness}",
                    range=error_builder.decorator_kwarg_value_range(kwarg="max_staleness"),
                    raise_error=ValueError,
                    code="13",
                )
        registry = CURRENT_FEATURE_REGISTRY.get()
        registry_features = registry.get_feature_sets()
        previous_features_class = registry_features.get(namespace, None)

        if (
            previous_features_class is not None
            and not notebook.is_notebook()
            and source_info.tree is not None
            and previous_features_class.__chalk_source_info__.tree is not None
            and (
                source_info.filename != previous_features_class.__chalk_source_info__.filename
                or source_info.tree.lineno != previous_features_class.__chalk_source_info__.tree.lineno
                or source_info.class_source != previous_features_class.__chalk_source_info__.class_source
            )
        ):
            error_builder.add_diagnostic(
                message=(
                    f"Feature class '{previous_features_class.__name__}' is defined twice: "
                    f"once in '{c.__module__}' and once in '{previous_features_class.__module__}'."
                ),
                code="14",
                label="duplicate class",
                range=error_builder.decorator_kwarg_value_range("name") or error_builder.class_definition_range(),
                raise_error=ValueError,
            )

        if (
            notebook.is_notebook()
            and previous_features_class is not None
            and notebook.is_defined_in_module(previous_features_class)
        ):
            # Not generating an LSP here because we're in a notebook anyway
            # TODO: See if we can pretty-print lsp errors in notebooks, at which point we can generate one that points to the old feature class
            raise ValueError(
                f"Cannot re-define feature class '{previous_features_class.__name__}' in a notebook: it was previously defined in '{previous_features_class.__module__}'."
            )

        updated_class = _process_class(
            cls=c,
            source_info=source_info,
            error_builder=error_builder,
            owner=owner,
            tags=ensure_tuple(tags),
            etl_offline_to_online=etl_offline_to_online,
            max_staleness=max_staleness,
            namespace=namespace,
            singleton=singleton,
        )
        assert is_features_cls(updated_class)

        registry.add_feature_set(updated_class)
        return cast(Type[T], updated_class)

    # See if we're being called as @features or @features().
    if cls is None:
        # We're called with parens.
        return wrap

    # We're called as @features without parens.
    return wrap(cls)


def _discover_feature(
    fs: Sequence[Feature],
    name: str,
    *conditions: Callable[[Feature], bool],
) -> Optional[Feature]:
    """
    Parameters
    ----------
    fs
        The features to search
    name
        Used for error messages
    conditions
        Tested in order. The first feature that matches _any_ condition is returned.
    """
    for cond in conditions:
        filtered_features = [c for c in fs if cond(c)]

        if len(filtered_features) == 1:
            return filtered_features[0]

        if len(filtered_features) > 1:
            assert filtered_features[0].features_cls is not None
            representative = filtered_features[0]
            assert representative.features_cls is not None
            b = representative.lsp_error_builder.add_diagnostic(
                message=(
                    f"Multiple {name} features are not supported in {representative.features_cls.__name__}: "
                    + ", ".join(f"{representative.features_cls.__name__}.{x.name}" for x in filtered_features)
                ),
                code="51",
                label=f"duplicate {name} feature",
                range=representative.lsp_error_builder.property_range(representative.attribute_name),
                raise_error=ValueError,
            )

            for ff in filtered_features[1:]:
                b.with_range(
                    range=ff.lsp_error_builder.property_range(ff.attribute_name),
                    label=f"duplicate {name} feature",
                )

    return None


def _setattr_fn(bidirectional_alias: Mapping[str, str]):
    def setattr(self: object, key: str, value: object):
        if key in bidirectional_alias:
            super(self.__class__, self).__setattr__(bidirectional_alias[key], value)
        return super(self.__class__, self).__setattr__(key, value)

    setattr.__name__ = "__setattr__"
    return setattr


def _getattribute_fn(self: object, attribute_name: str):
    # If calling getattr() on an instance for a feature name,
    # do NOT return a class-level FeatureWrapper.
    # Instead, raise an attribute error
    o = object.__getattribute__(self, attribute_name)
    if isinstance(o, FeatureWrapper):
        raise AttributeError(
            f"Feature '{attribute_name}' is not defined on this instance of class'{type(self).__name__}'"
        )
    return o


_getattribute_fn.__name__ = "__getattribute__"


def _init_fn(
    additional_inits: frozenset[str],
    alias_from_to: Mapping[str, str],
):
    def _init(self: Features, /, *args: object, **kwargs: object):
        for feat, val in zip(self.features, args):
            if feat.attribute_name in kwargs:
                raise ValueError(
                    f"Argument '{feat.attribute_name}' cannot be specified as both a positional and keyword argument"
                )
            kwargs[feat.attribute_name] = val

        for from_, to in alias_from_to.items():
            if from_ in kwargs and to in kwargs:
                raise ValueError(
                    f"The features '{from_}' and '{to}' are aliases of each other. Only one can be specified, but both were given."
                )
            if to in kwargs:
                kwargs[from_] = kwargs[to]
        assert self.__chalk_ts__ is not None
        ts_feature_name = self.__chalk_ts__.attribute_name

        for k, v in kwargs.items():
            cls_field = getattr(self.__class__, k, None)
            if type(cls_field) is not FeatureWrapper and k not in additional_inits and k != ts_feature_name:
                raise TypeError(f"{self.__class__.__name__}.__init__() got an unexpected keyword argument '{k}'")
            setattr(self, k, v)

    _init.__name__ = "__init__"
    return _init


def _parse_tags(ts: str) -> List[str]:
    ts = re.sub(",", " ", ts)
    ts = re.sub(" +", " ", ts.strip())
    return [xx.strip() for xx in ts.split(" ")]


def _get_field(
    cls: Type,
    error_builder: FeatureClassErrorBuilder,
    annotation_name: str,
    comments: Mapping[str, str],
    class_owner: Optional[str],
    class_tags: Optional[Tuple[str, ...]],
    class_etl_offline_to_online: bool,
    class_max_staleness: timedelta,
    namespace: str,
    is_singleton: bool,
) -> Feature:
    # Return a Field object for this field name and type.  ClassVars and
    # InitVars are also returned, but marked as such (see f._field_type).
    # default_kw_only is the value of kw_only to use if there isn't a field()
    # that defines it.

    # If the default value isn't derived from Field, then it's only a
    # normal default value.  Convert it to a Field().
    last_value = LSPErrorBuilder.lsp
    LSPErrorBuilder.lsp = False
    default = getattr(cls, annotation_name, ...)
    LSPErrorBuilder.lsp = last_value

    if isinstance(default, Feature):
        # The feature was set like x: int = feature(...)
        f = default
        if f.unversioned_attribute_name is None and not f.is_autogenerated:
            f.unversioned_attribute_name = annotation_name
        if f.version is not None:
            f.version.base_name = f.name if f.is_name_set() else annotation_name
            f.name = f.version.name_for_version(f.version.default)
        if f.is_name_set():
            if "." in f.name:
                error_builder.add_diagnostic(
                    message=(
                        f"Custom feature names cannot contain a dot, but the feature '{f.name}' on the class '{cls.__name__}' includes a dot. You might consider using a has-one feature instead."
                    ),
                    code="75",
                    label="dotted name",
                    range=error_builder.property_value_kwarg_range(annotation_name, kwarg="name")
                    or error_builder.property_range(annotation_name),
                    raise_error=ValueError,
                )
            elif " " in f.name:
                error_builder.add_diagnostic(
                    message=(
                        f"Custom feature names cannot contain spaces, but the feature '{f.name}' on the class '{cls.__name__}' includes a space."
                    ),
                    code="75",
                    label="name with space",
                    range=error_builder.property_value_kwarg_range(annotation_name, kwarg="name")
                    or error_builder.property_range(annotation_name),
                    raise_error=ValueError,
                )

    elif isinstance(default, Windowed):
        # The feature was set like x: Windowed[int] = windowed()
        # Convert it to a Feature
        try:
            f = default._to_feature(bucket=None)
        except Exception as e:
            error_builder.add_diagnostic(
                message=f"Invalid window found for feature '{namespace}.{annotation_name}'. {e}",
                label="invalid duration",
                code="18",
                range=error_builder.property_value_range(annotation_name),
                raise_error=ValueError,
            )
            f = default
        if f.version is not None:
            f.version.base_name = f.name if f.is_name_set() else annotation_name
            f.name = f.version.name_for_version(f.version.default)

    else:
        underscore_expression = None
        # The feature was not set explicitly
        if isinstance(default, types.MemberDescriptorType):
            # This is a field in __slots__, so it has no default value.
            default = ...
        if isinstance(default, Underscore):
            underscore_expression = default
            default = ...

        f = Feature(
            name=annotation_name,
            namespace=namespace,
            default=default,
            underscore_expression=underscore_expression,
        )

    # Only at this point do we know the name and the type.  Set them.
    f.namespace = namespace
    if not f.is_typ_set():
        f.typ = ParsedAnnotation(cls, annotation_name)

    f.features_cls = cls
    f.attribute_name = annotation_name
    if not hasattr(f, "name"):
        f.name = annotation_name
    f.is_singleton = is_singleton
    _process_field(
        f=f,
        comments=comments,
        class_owner=class_owner,
        class_tags=class_tags,
        class_etl_offline_to_online=class_etl_offline_to_online,
        class_max_staleness=class_max_staleness,
        error_builder=error_builder,
    )
    return f


def _process_field(
    f: Feature,
    comments: Mapping[str, str],
    class_owner: Optional[str],
    class_tags: Optional[Tuple[str, ...]],
    class_etl_offline_to_online: bool,
    class_max_staleness: timedelta,
    error_builder: FeatureClassErrorBuilder,
) -> Feature:
    comment_for_feature = comments.get(f.attribute_name)
    comment_based_description = None
    comment_based_owner = None
    comment_based_tags = None

    if comment_for_feature is not None:
        comment_lines = []
        for line in comment_for_feature.splitlines():
            stripped_line = line.strip()
            if stripped_line.startswith(":owner:"):
                comment_based_owner = stripped_line.removeprefix(":owner:").strip()
            elif stripped_line.startswith(":tags:"):
                parsed = _parse_tags(stripped_line.removeprefix(":tags:"))
                if len(parsed) > 0:
                    if comment_based_tags is None:
                        comment_based_tags = parsed
                    else:
                        comment_based_tags.extend(parsed)
            else:
                comment_lines.append(line)

        comment_based_description = "\n".join(comment_lines)

    if f.description is None and comment_based_description:
        f.description = comment_based_description

    if comment_based_tags is not None:
        if f.tags is None:
            f.tags = comment_based_tags
        else:
            f.tags.extend(comment_based_tags)

    if class_tags is not None:
        if f.tags is None:
            f.tags = list(class_tags)
        else:
            f.tags.extend(class_tags)

    if f.owner is not None and comment_based_owner is not None:
        assert f.features_cls is not None
        error_builder.add_diagnostic(
            message=(
                f"Owner for feature '{f.name}' on class '{f.features_cls.__name__}' "
                f"specified both on the feature and in the comment. Please use only one of these two."
            ),
            code="15",
            label="second declaration",
            range=error_builder.property_value_kwarg_range(f.attribute_name, kwarg="owner")
            or error_builder.property_range(f.attribute_name),
            raise_error=ValueError,
        )

    elif f.owner is None:
        f.owner = comment_based_owner or class_owner

    # The attribute is not defined if the feature intends to use the class default
    if not hasattr(f, "max_staleness"):
        f.max_staleness = class_max_staleness

    # Using the private variable because the etl_offline_to_online is a read-only property
    if not hasattr(f, "etl_offline_to_online"):
        f.etl_offline_to_online = class_etl_offline_to_online
    return f


def _repr_fn(self: Features):
    parts: list[str] = []
    for f in self.features:
        val = getattr(self, f.attribute_name, MISSING)
        if val is not MISSING:
            parts.append(f"{f.attribute_name}={val!r}")
    return f"{self.__class__.__name__}(" + ", ".join(parts) + ")"


_repr_fn.__name__ = "__repr__"


def _eq_fn(self: Features, other: object):
    if not isinstance(other, type(self)):
        return NotImplemented
    return all(
        getattr(self, f.attribute_name, MISSING) == getattr(other, f.attribute_name, MISSING) for f in self.features
    )


_eq_fn.__name__ = "__eq__"


def _len_fn(self: Features):
    count = 0
    for f in self.features:
        if not f.no_display and hasattr(self, f.attribute_name):
            count += 1
    return count


_len_fn.__name__ = "__len__"


def _items_fn(self: Features):
    for f in self.features:
        if not f.no_display and hasattr(self, f.attribute_name):
            yield f.fqn, getattr(self, f.attribute_name)


_items_fn.__name__ = "__items__"


def _iter_fn(self: Features):
    for f in self.features:
        if (
            hasattr(self, f.attribute_name)
            and type(f).__name__ == "Feature"
            and not f.is_has_one
            and not f.is_has_many
            and not f.no_display
        ):
            yield f.fqn, getattr(self, f.attribute_name)


_iter_fn.__name__ = "__iter__"


def _parse_annotation_comments(source_info: ClassSource, cls_annotations: Dict[str, Any]) -> Mapping[str, str]:
    if source_info.tree is None or source_info.dedent_source is None:
        return {}

    source_lines = source_info.dedent_source.splitlines()
    """ Get rid of the decorator, if exists (python 3.8 won't have one)"""
    for i, line in enumerate(source_lines):
        if line.lstrip().startswith("class "):
            source_lines = source_lines[i:]
            break

    comments_for_annotations: Dict[str, str] = {}

    for stmt in source_info.tree.body:
        if isinstance(stmt, ast.AnnAssign) and isinstance(stmt.target, ast.Name):
            line = stmt.lineno - source_info.tree.lineno - 1
            comments: List[str] = []
            while line >= 0 and source_lines[line].strip().startswith("#"):
                comment = source_lines[line].strip().lstrip("#").strip()
                comments.insert(0, comment)
                line -= 1

            if len(comments) > 0:
                comments_for_annotations[stmt.target.id] = textwrap.dedent("\n".join(comments))

    """Attach the same comments to windowed features as well"""
    for annotation, feature_type in cls_annotations.items():
        if annotation in comments_for_annotations and isinstance(feature_type, Windowed):
            for bucket_size in feature_type.buckets_seconds:
                pseudofeature_name = get_name_with_duration(annotation, bucket_size)
                comments_for_annotations[pseudofeature_name] = comments_for_annotations[annotation]

    return comments_for_annotations


CHALK_SINGLETON_VALUE = 111


def _process_group_by_windowed(
    cls: Type,
    namespace: str,
    attribute_name: str,
    gbw: GroupByWindowed,
    cls_annotations: Dict[str, Any],
    error_builder: FeatureClassErrorBuilder,
) -> Feature:
    if gbw._name is None:
        gbw._name = gbw._name if gbw._name is not None else attribute_name

    gbw._namespace = namespace

    if gbw._dtype is not None:
        gbw._kind = pyarrow_to_primitive(gbw._dtype, name=attribute_name)
        if gbw._kind not in (int, float):
            error_builder.add_diagnostic(
                message=(
                    f"group_by_window feature '{namespace}.{attribute_name}' has an invalid dtype '{gbw._dtype}'. "
                    "Please use a numeric integer or float dtype, like 'int64' or 'float64'."
                ),
                label="invalid dtype",
                range=error_builder.property_value_kwarg_range(attribute_name, kwarg="dtype")
                or error_builder.property_value_range(attribute_name)
                or error_builder.property_range(attribute_name),
                code="16",
            )
    else:
        gbw._dtype = pa.float32()
        gbw._kind = float

    return Feature(
        name=gbw._name,
        attribute_name=attribute_name,
        namespace=namespace,
        features_cls=cls,
        typ=gbw._kind,
        version=None,
        default_version=1,
        description=gbw._description,
        owner=gbw._owner,
        tags=gbw._tags,
        primary=False,
        default=gbw._default,
        underscore_expression=gbw._expression,
        max_staleness=None,
        cache_strategy=CacheStrategy.ALL,
        etl_offline_to_online=None,
        encoder=None,
        decoder=None,
        pyarrow_dtype=gbw._dtype,
        join=None,
        is_feature_time=False,
        is_autogenerated=False,
        validations=None,
        all_validations=None,
        window_durations=tuple(),  # Not here, but on group by windowed
        window_duration=None,
        no_display=False,
        offline_ttl=None,
        last_for=None,
        hook=None,
        is_distance_pseudofeature=False,
        is_pseudofeature=False,
        window_materialization=gbw._materialization,
        group_by_windowed=gbw,
    )


_skip_attrs = {
    "__dict__",
    "__class__",
    "__annotations__",
    "__delattr__",
    "__dir__",
    "__doc__",
    "__eq__",
    "__format__",
    "__ge__",
    "__getattribute__",
    "__gt__",
    "__hash__",
    "__init__",
}


def _process_class(
    cls: Type[T],
    source_info: ClassSource,
    error_builder: FeatureClassErrorBuilder,
    owner: Optional[str],
    tags: Tuple[str, ...],
    etl_offline_to_online: bool,
    max_staleness: timedelta,
    namespace: str,
    singleton: bool,
) -> Type[T]:
    raw_cls_annotations = cls.__dict__.get("__annotations__", {})

    alias_from_to: Dict[str, str] = {}
    additional_inits: list[str] = []

    materialized_windows: list[Feature] = []
    group_by_materialized_windows: list[Feature] = []
    expression_windows: List[Feature] = []
    cls_annotations: Dict[str, Any] = {}
    for name, annotation in raw_cls_annotations.items():
        if name in ("features", "namespace", "items", "is_near"):
            error_builder.add_diagnostic(
                message=f"Feature '{name}' on class '{cls.__name__}' uses a reserved name.",
                label=f"reserved name '{name}'",
                range=error_builder.property_range(name),
                raise_error=TypeError,
                code="16",
            )

        if isinstance(annotation, Windowed) or (isinstance(annotation, str) and "Windowed" in annotation):
            # NOTE: For Windowed resolvers, both the Annotation and the value are instances of
            # Windowed, unlike normal features whose annotation is the underlying type, and the
            # value is an instance of FeatureWrapper. So both `annotation` and `wind` should be
            # instances of Windowed.
            #
            # In the future, we should use a subclass of Windowed, rather than an instance, for
            # the type annotation, similar to what we do for Features.
            if isinstance(annotation, str) and "Windowed" in annotation:
                try:
                    annotation = parse_quoted_window_feature(annotation, cls.__module__)
                except Exception as e:
                    error_builder.add_diagnostic(
                        message=(
                            f"Quoted Windowed feature type annotation '{annotation}' for '{namespace}.{name}' could not be parsed: {e}."
                            f"To create a windowed feature with quotes for typing, use "
                            f'\'{name}: "Windowed[int]" = windowed("10m", ...)\''
                        ),
                        label="missing windowed(...) call",
                        range=error_builder.property_range(name),
                        raise_error=TypeError,
                        code="17",
                    )
            assert isinstance(annotation, Windowed), f"failed to parse annotation {annotation} as Windowed"

            wind = getattr(cls, name, None)
            if wind is None or not isinstance(wind, Windowed):
                assert annotation._kind is not None
                error_builder.add_diagnostic(
                    message=(
                        f"Windowed feature '{namespace}.{name}' is missing windows. "
                        f"To create a windowed feature, use "
                        f"'{name}: Windowed[{annotation.kind.__name__}] = windowed(\"10m\", ...)'"
                    ),
                    label="missing windowed(...) call",
                    range=error_builder.property_range(name),
                    raise_error=TypeError,
                    code="17",
                )

            wind.kind = annotation.kind
            if wind._name is None:
                wind._name = name
            annotation._buckets = wind._buckets
            if len(wind._buckets) == 0:
                error_builder.add_diagnostic(
                    message=(
                        f"Windowed feature '{namespace}.{name}' does not have any window durations. "
                        f"To create a windowed feature, use "
                        f'\'{name}: Windowed[{annotation.kind.__name__}] = windowed("1h", "1d"))\''
                    ),
                    label="missing window durations",
                    range=error_builder.property_value_range(name) or error_builder.property_range(name),
                    code="777",
                )

            if isinstance(wind._materialization, dict):

                def get_mat_range():
                    return (
                        error_builder.property_value_kwarg_range(name, kwarg="materialization")
                        or error_builder.property_value_range(name)
                        or error_builder.property_range(name)
                    )

                if wind._expression is None:
                    error_builder.add_diagnostic(
                        message=(
                            f"Windowed feature '{namespace}.{name}' has a materialization, but no expression. "
                            f"To create a materialized windowed feature, use "
                            f"'{name}: Windowed[{annotation.kind.__name__}] = windowed(\"10m\", materialization={wind._materialization}, expression=_.your_dataframe_feature[_.field_to_agg].sum())'"
                        ),
                        label="materialization config",
                        range=get_mat_range(),
                        code="177",
                    )

                bucket_duration_raw = wind._materialization.get("bucket_duration", None)
                if bucket_duration_raw is not None:
                    try:
                        wind._materialization["bucket_duration"] = parse_chalk_duration(bucket_duration_raw)
                    except ValueError as e:
                        error_builder.add_diagnostic(
                            message=(
                                f"Windowed feature '{namespace}.{name}' has an invalid `bucket_duration`. {e.args[0]}"
                            ),
                            label="invalid duration",
                            range=get_mat_range(),
                            code="18",
                        )

                validated_bucket_durations: dict[Duration, list[Duration]] = collections.defaultdict(list)
                for d, windows_or_window in wind._materialization.get("bucket_durations", {}).items():
                    try:
                        d = parse_chalk_duration(d)
                    except ValueError as e:
                        error_builder.add_diagnostic(
                            message=(
                                f"Windowed feature '{namespace}.{name}' has an invalid key in `bucket_durations`. {e.args[0]}"
                            ),
                            label="invalid duration",
                            range=get_mat_range(),
                            code="18",
                            raise_error=ValueError,
                        )

                    for_windows = ensure_tuple(windows_or_window)
                    for w in for_windows:
                        try:
                            w = parse_chalk_duration(w)
                        except ValueError as e:
                            error_builder.add_diagnostic(
                                message=(
                                    f"Windowed feature '{namespace}.{name}' has an invalid key in `bucket_durations`. {e.args[0]}"
                                ),
                                label="invalid duration",
                                range=get_mat_range(),
                                code="18",
                                raise_error=ValueError,
                            )
                        validated_bucket_durations[d].append(w)

                wind._materialization["bucket_durations"] = validated_bucket_durations
                if bucket_duration_raw is None and not validated_bucket_durations:
                    error_builder.add_diagnostic(
                        message=(
                            f"Windowed feature '{namespace}.{name}' has a materialization, but no 'bucket_duration'. "
                            "Please provide a 'bucket_duration', like '1h', in the dictionary under the keyword "
                            "argument 'materialization'."
                        ),
                        label="missing 'bucket_duration' key",
                        range=get_mat_range(),
                        code="188",
                    )

            for bucket in wind._buckets:
                # Make pseudo-features for each bucket of the window
                try:
                    bucket_seconds = parse_chalk_duration_s(bucket)
                except ValueError as e:
                    # set bucket_seconds to anything so that we can continue accumulating errors
                    # in the class.
                    bucket_seconds = 60
                    error_builder.add_diagnostic(
                        message=f"Invalid window found for feature '{namespace}.{name}'. {e.args[0]}",
                        label="invalid duration",
                        range=error_builder.property_value_range(name),
                        code="18",
                    )

                try:
                    feat = wind._to_feature(bucket=bucket_seconds)
                except ValueError as e:
                    error_builder.add_diagnostic(
                        message=f"Invalid window found for feature '{namespace}.{name}'. {e.args[0]}",
                        label="invalid duration",
                        range=error_builder.property_value_range(name),
                        code="18",
                    )
                    continue

                if feat.window_materialization is not None:
                    materialized_windows.append(feat)
                elif feat.underscore_expression is not None:
                    expression_windows.append(feat)

                # For the pseudo-features, which track an individual bucket,
                # the correct annotation is the underlying annotation, not
                # Windowed[underlying], since it's only one value
                cls_annotations[feat.name] = wind.kind
                setattr(cls, feat.name, feat)
                if isinstance(bucket, str):  # pyright: ignore [reportUnnecessaryIsInstance] -- user provided
                    alias = f"{name}_{bucket}"
                    additional_inits.append(alias)
                    alias_from_to[
                        get_name_with_duration(
                            name_or_fqn=name,
                            duration="all" if bucket == "infinity" or bucket == "all" else bucket_seconds,
                        )
                    ] = alias

        windowed_group_by = getattr(cls, name, None)
        if isinstance(annotation, DataFrameMeta) and isinstance(windowed_group_by, GroupByWindowed):
            feat = _process_group_by_windowed(
                cls=cls,
                namespace=namespace,
                attribute_name=name,
                gbw=windowed_group_by,
                cls_annotations=cls_annotations,
                error_builder=error_builder,
            )
            group_by_materialized_windows.append(feat)
            setattr(cls, name, feat)
            kind = windowed_group_by._kind
            cls_annotations[name] = kind
        else:
            cls_annotations[name] = annotation

    _globals = sys.modules[cls.__module__].__dict__ if cls.__module__ in sys.modules else {}

    for name, member in inspect.getmembers(cls):
        if name in _skip_attrs:
            continue
        if name not in cls_annotations and isinstance(member, Windowed):
            error_builder.add_diagnostic(
                range=error_builder.property_range(name),
                message=f"Windowed feature '{namespace}.{name}' is missing an annotation, like 'Windowed[str]'",
                label="missing annotation",
                raise_error=TypeError,
                code="20",
            )

        # Feature times that weren't annotated.
        if name not in cls_annotations and isinstance(member, Feature):
            # All feature types need annotations, except for datetimes, which we can automatically infer
            if member._is_feature_time:
                # We must read the private variable to avoid parsing the annotation,
                # which might contain forward references that are not yet loaded
                cls_annotations[name] = datetime
            else:
                error_builder.add_diagnostic(
                    code="18",
                    range=error_builder.property_range(name),
                    message=f"Feature '{namespace}.{name}' is missing an annotation. Please add one, like '{name}: str = ...'",
                    label="missing annotation",
                    raise_error=TypeError,
                )
        if name not in cls_annotations and isinstance(member, str) and not name.startswith("__"):
            error_builder.add_diagnostic(
                code="18",
                range=error_builder.property_range(name),
                message=(
                    f"Feature '{namespace}.{name}' is missing an annotation. Did you use an equals sign instead of a colon? "
                    f"If so, please change your feature definition to '{name}: {member}'"
                ),
                label="missing annotation",
                raise_error=TypeError,
            )

        if name not in cls_annotations and isinstance(member, Underscore):
            error_builder.add_diagnostic(
                code="18",
                range=error_builder.property_range(name),
                message=f"Underscore feature '{namespace}.{name}' is missing an annotation. Please add one, like '{name}: int = ...'",
                label="missing annotation",
                raise_error=TypeError,
            )

        # catch malformed features like num_transactions: int = windowed("2h") with no Windowed type
        if isinstance(member, Windowed) and not isinstance(cls_annotations.get(name), Windowed):
            error_builder.add_diagnostic(
                message=(
                    f"Feature '{namespace}.{name}' is marked as 'windowed()', "
                    f"but also needs to be marked as a Windowed type. "
                    f"Please add one, like '{name}: Windowed[int] = ...'"
                ),
                range=error_builder.annotation_range(name) or error_builder.property_range(name),
                label="missing Windowed[...]",
                raise_error=TypeError,
                code="19",
            )

    cls.__annotations__ = cls_annotations
    del cls_annotations  # unused; set cls.__annotations__ directly

    set_new_attribute(cls=cls, name="__chalk_is_singleton__", value=singleton)
    set_new_attribute(cls=cls, name="__chalk_error_builder__", value=error_builder)
    set_new_attribute(cls=cls, name="__chalk_source_info__", value=source_info)
    set_new_attribute(cls=cls, name="__chalk_materialized_windows__", value=materialized_windows)
    set_new_attribute(cls=cls, name="__chalk_group_by_materialized_windows__", value=group_by_materialized_windows)
    set_new_attribute(cls=cls, name="__chalk_expression_windows__", value=expression_windows)

    cls_fields: List[Feature] = []

    if singleton:
        f = Feature(
            primary=True,
            attribute_name="__chalk_singleton_id__",
            name="__chalk_singleton_id__",
            default=CHALK_SINGLETON_VALUE,
            namespace=namespace,
            typ=int,
            pyarrow_dtype=pa.uint8(),
            max_staleness=None,
            etl_offline_to_online=False,
            is_autogenerated=True,
            no_display=True,
        )
        f.is_singleton = True
        set_new_attribute(cls=cls, name="__chalk_singleton_id__", value=f)
        cls.__annotations__[f.attribute_name] = int

    def __chalk_primary__(_: Type[Features]):
        return _discover_feature(
            cls_fields,
            "primary",
            lambda q: q._primary is True,
            lambda q: q.typ.is_primary(),
            lambda q: (
                q.name == "id" and not q.has_resolved_join and not q._is_feature_time and not q.typ.is_feature_time()
            ),
        )

    set_new_attribute(cls, "__chalk_primary__", value=classproperty(__chalk_primary__, cached=True))

    def __chalk_ts__(cl: Type[Features]) -> Feature:
        # Not using `f.is_feature_time` as that would create an infinite recursion, since
        # `.is_feature_time` accesses `__chalk_ts__`
        ts_feature: Optional[Feature] = _discover_feature(
            cls_fields,
            "feature time",
            lambda q: q._is_feature_time is True,
            lambda q: q.typ.is_feature_time(),
            lambda q: q.name == "ts" and not q.has_resolved_join and not q._primary and not q.typ.is_primary(),
        )
        if ts_feature is None:
            return unwrap_feature(getattr(cl, GENERATED_OBSERVED_AT_NAME))
        return ts_feature

    set_new_attribute(
        cls=cls,
        name="__chalk_ts__",
        value=classproperty(__chalk_ts__, cached=True),
    )

    def __chalk_observed_at__(cls: Type[Features]) -> FeatureWrapper:
        ts_feature: Optional[Feature] = _discover_feature(
            cls_fields,
            "feature time",
            # Not using `f.is_feature_time` as that would create an infinite recursion, since
            # `.is_feature_time` accesses `__chalk_ts__` which can access this function
            lambda f: f._is_feature_time is True,
            lambda f: f.typ.is_feature_time(),
            lambda f: f.name == "ts" and not f.has_resolved_join and not f._primary and not f.typ.is_primary(),
        )
        if ts_feature is not None:
            if ts_feature.attribute_name != GENERATED_OBSERVED_AT_NAME:
                error_builder.add_diagnostic(
                    message=f"Object {cls.__name__} has no attribute '{GENERATED_OBSERVED_AT_NAME}",
                    label="missing attribute",
                    range=error_builder.property_range(ts_feature.attribute_name),
                    raise_error=AttributeError,
                    code="26",
                )

        # If the timestamp feature is still none, then synthesize one on first use
        ts_feature = feature_time()
        assert ts_feature is not None
        ts_feature.name = GENERATED_OBSERVED_AT_NAME
        ts_feature.attribute_name = GENERATED_OBSERVED_AT_NAME
        ts_feature.unversioned_attribute_name = None
        ts_feature.namespace = cls.__chalk_namespace__
        ts_feature.features_cls = cls
        ts_feature.is_autogenerated = True
        cls.__annotations__[GENERATED_OBSERVED_AT_NAME] = datetime
        _process_field(
            f=ts_feature,
            error_builder=error_builder,
            comments={},
            class_owner=cls.__chalk_owner__,
            class_tags=tuple(cls.__chalk_tags__),
            class_etl_offline_to_online=cls.__chalk_etl_offline_to_online__,
            class_max_staleness=cls.__chalk_max_staleness__,
        )

        return FeatureWrapper(ts_feature)

    set_new_attribute(
        cls=cls,
        name=GENERATED_OBSERVED_AT_NAME,
        value=classproperty(__chalk_observed_at__, cached=True, bind_to_instances=False),
    )

    def __features__(cl: Type[Features]) -> List[Feature]:
        fs = list(cls_fields)
        if cl.__chalk_ts__ not in fs:
            assert cl.__chalk_ts__ is not None
            fs.append(cl.__chalk_ts__)
        return fs

    set_new_attribute(cls=cls, name="features", value=classproperty(__features__, cached=True))
    set_new_attribute(cls=cls, name="__str__", value=classmethod(lambda _: namespace))
    set_new_attribute(cls=cls, name="__chalk_features_raw__", value=cls_fields)
    set_new_attribute(cls=cls, name="__chalk_is_loaded_from_notebook__", value=False)
    set_new_attribute(cls=cls, name="__chalk_notebook_feature_expressions__", value=dict())
    set_new_attribute(cls=cls, name="__repr__", value=_repr_fn)
    set_new_attribute(cls=cls, name="__eq__", value=_eq_fn)
    set_new_attribute(cls=cls, name="__hash__", value=None)
    set_new_attribute(cls=cls, name="__iter__", value=_iter_fn)
    set_new_attribute(cls=cls, name="items", value=_items_fn)
    set_new_attribute(cls=cls, name="namespace", value=namespace)
    set_new_attribute(cls=cls, name="__chalk_namespace__", value=namespace)
    set_new_attribute(cls=cls, name="__chalk_owner__", value=owner)
    set_new_attribute(cls=cls, name="__chalk_tags__", value=list(tags))
    set_new_attribute(cls=cls, name="__chalk_max_staleness__", value=max_staleness)
    set_new_attribute(cls=cls, name="__is_features__", value=True)
    set_new_attribute(cls=cls, name="__len__", value=_len_fn)
    set_new_attribute(cls=cls, name="__getattribute__", value=_getattribute_fn)
    set_new_attribute(cls=cls, name="__setattr__", value=classmethod(_class_setattr))
    set_new_attribute(
        cls=cls,
        name="__chalk_etl_offline_to_online__",
        value=etl_offline_to_online,
    )
    # Moving this line lower causes all kinds of problems.
    cls = classproperty_support(cls)

    comments = {}
    try:
        # If we pass this function something that isn't a class, it could raise
        comments = _parse_annotation_comments(source_info, cls.__annotations__)
    except:
        pass

    # Parse the fields after we have the correct `cls` set
    cls_fields.extend(
        _get_field(
            cls=cls,
            error_builder=error_builder,
            annotation_name=name,
            comments=comments,
            class_owner=owner,
            class_tags=tags,
            class_etl_offline_to_online=etl_offline_to_online,
            class_max_staleness=max_staleness,
            namespace=namespace,
            is_singleton=singleton,
        )
        for name in cls.__annotations__
    )
    for f in tuple(cls_fields):
        if singleton and f.primary and f.name != "__chalk_singleton_id__":
            error_builder.add_diagnostic(
                message=(
                    f"The singleton feature class '{namespace}' includes a feature '{f.name}' that is primary. "
                    f"Please remove the feature '{f.fqn}' "
                    f"or remove the singleton keyword argument to the feature class '{namespace}'."
                ),
                range=error_builder.property_range(f.attribute_name),
                label="primary feature",
                raise_error=ValueError,
                code="27",
            )

        if f.version is None:
            continue

        alias_from_to[f.attribute_name] = f"{f.attribute_name}_v{f.version.default}"

        if f.version.explicitly_enumerated:
            for i, mapped_feature in f.version.reference.items():
                f_i = copy.copy(mapped_feature)
                f_i.namespace = namespace

                f_i.version = VersionInfo(
                    version=i,
                    maximum=f.version.maximum,
                    default=f.version.default,
                    reference=f.version.reference,
                    base_name=getattr(f_i, "name", None) or f.version.base_name,
                    explicitly_enumerated=False,
                )
                cls_cast = cast(Type, cls)
                if not f_i.is_typ_set():
                    f_i.typ = ParsedAnnotation(cls_cast, f.name)

                f_i.features_cls = cls_cast
                f_i.attribute_name = f"{f.attribute_name}_v{i}"
                if not hasattr(f_i, "name"):
                    f_i.name = f_i.version.name_for_version(i)
                f_i.unversioned_attribute_name = f.attribute_name
                f_i.is_singleton = f.is_singleton
                _process_field(
                    f=f_i,
                    comments=comments,
                    class_owner=owner,
                    class_tags=tags,
                    class_etl_offline_to_online=etl_offline_to_online,
                    class_max_staleness=max_staleness,
                    error_builder=error_builder,
                )
                f.version.reference[i] = f_i
                # The default feature already exists.
                f_i.no_display = i == f.version.default
            continue

        for i in range(1, f.version.maximum + 1):
            f_i = copy.copy(f)
            f_i.name = f.version.name_for_version(i)
            f_i.unversioned_attribute_name = f.attribute_name
            f_i.attribute_name = f"{f.attribute_name}_v{i}"
            f_i.version = VersionInfo(
                version=i,
                maximum=f.version.maximum,
                default=f.version.default,
                reference=f.version.reference,
                base_name=f.version.base_name,
                explicitly_enumerated=False,
            )
            f.version.reference[i] = f_i
            cls_fields.append(f_i)

            # The default feature already exists.
            f_i.no_display = i == f.version.default

            if f_i.attribute_name in cls.__annotations__:
                assert f_i.features_cls is not None
                error_builder.add_diagnostic(
                    message=(
                        f"The class '{f_i.features_cls.__name__}' "
                        f"has an existing annotation '{f_i.attribute_name}' "
                        "that collides with a versioned feature. Please remove the existing "
                        "annotation, or lower the version."
                    ),
                    range=error_builder.property_range(f_i.attribute_name),
                    label="invalid name",
                    raise_error=ValueError,
                    code="21",
                )
            cls.__annotations__[f_i.attribute_name] = cls.__annotations__[f.attribute_name]

    set_new_attribute(
        cls=cls,
        name="__setattr__",
        value=_setattr_fn(
            bidirectional_alias={**{v: k for k, v in alias_from_to.items()}, **alias_from_to},
        ),
    )

    set_new_attribute(
        cls=cls,
        name="__init__",
        value=_init_fn(
            additional_inits=frozenset(additional_inits),
            alias_from_to=alias_from_to,
        ),
    )

    for f in cls_fields:
        assert f.attribute_name is not None
        f.features_cls = cast(Type[Features], cls)
        # Wrap all class features with FeatureWrapper
        setattr(cls, f.attribute_name, FeatureWrapper(f))

        if f.hook:
            f.hook(cast(Type[Features], cls))
    set_new_attribute(cls=cls, name="__chalk_feature_set__", value=True)
    return cls


def _class_setattr(
    cls: Type[Features],
    key: str,
    value: Any,
):
    from chalk.features.feature_set import FeatureSetBase

    # Handle inline feature definitions in notebooks
    if (
        (key.startswith("__") and key.endswith("__"))
        or key == "features"
        or key == "namespace"
        or isinstance(value, FeatureWrapper)
    ):
        # If it's a dunder, then  set it directly
        # If it's the literal 'features' or 'namespace', then set it directly -- Chalk reserves these names
        # If it's already a FeatureWrapper, then assume it was already constructed correctly, so set it directly
        type.__setattr__(cls, key, value)
        return
    f = None
    is_notebook_defined_feature = False
    if isinstance(value, Underscore):
        from chalk.df.ast_parser import parse_inline_setattr_annotation

        fqn = build_namespaced_name(name=f"{cls.namespace}.{key}")
        existing_feature = next((f for f in cls.features if f.fqn == fqn), None)

        is_notebook_expression: bool = notebook.is_notebook()
        if is_notebook_expression:
            # In notebooks, newly-defined underscores are stored separately, as "feature expressions".
            # These don't correspond to real features
            if existing_feature is not None:
                raise ValueError(f"Can't overwrite feature '{fqn}' because it already exists in the deployment source.")
            cls.__chalk_notebook_feature_expressions__[key] = NamedUnderscoreExpr(expr=value, fqn=fqn, short_name=key)
        else:
            typ = parse_inline_setattr_annotation(key)
            if typ is None:
                if existing_feature is None:
                    raise TypeError(f"Please define a type annotation for feature '{fqn}'")
                else:
                    parsed_annotation = ParsedAnnotation(underlying=existing_feature.typ.parsed_annotation)
            else:
                parsed_annotation = ParsedAnnotation(underlying=typ)

            if existing_feature is not None:
                existing_feature.typ = parsed_annotation
                existing_feature.underscore_expression = value
            else:
                f = Feature(
                    namespace=cls.namespace,
                    name=key,
                    attribute_name=key,
                    features_cls=cls,
                    typ=parsed_annotation,
                    underscore_expression=value,
                )
    elif isinstance(value, Feature):
        # Handle the case of `User.new_feat = feature(....)` in a notebook
        f = value
        if notebook.is_notebook():
            # Note: Can't really distinguish between features declared like this in a notebook cell
            # vs. features declared like this in code imported from a notebook cell (e.g. customer code)
            # Gonna treat both the same since this code only runs in a notebook context
            # (e.g. customers can't use this pattern as part of their deployment code)
            is_notebook_defined_feature = True
            existing_feature = next((ff for ff in cls.features if ff.unversioned_attribute_name == key), None)
            if notebook.is_defined_in_module(
                cls
            ) and key not in FeatureSetBase.__chalk_notebook_defined_feature_fields__.get(cls.namespace, set()):
                if existing_feature is not None:
                    raise ValueError(
                        f"Can't overwrite feature '{cls.namespace}.{key}' because it already exists in the deployment source."
                    )
            if hasattr(f, "max_staleness") and f.max_staleness.total_seconds() > 0:
                raise ValueError(
                    "Cannot set `max_staleness` on a notebook defined feature: persistence is not supported for notebook defined features"
                )
            # Get the type annotation
            parsed_annotation: Optional[ParsedAnnotation] = None
            if existing_feature is not None:
                parsed_annotation = ParsedAnnotation(underlying=existing_feature.typ.parsed_annotation)
            if f._typ is not None and f.typ._unparsed_underlying is not None:
                # If the feature already has a type annotation, use that
                parsed_annotation = f.typ
            else:
                from chalk.df.ast_parser import parse_inline_setattr_annotation

                typ = parse_inline_setattr_annotation(key)
                if typ is not None:
                    parsed_annotation = ParsedAnnotation(underlying=typ)
            if parsed_annotation is None:
                raise TypeError(
                    f"Please define a type annotation for feature '{cls.namespace}.{key}', like so: User.new_feature=feature(typ=str)"
                )
            # Set some attributes that usually get set when the feature is created inside the class definition
            f.attribute_name = key
            f.unversioned_attribute_name = key
            if getattr(f, "name", None) is None:
                f.name = key
            f.namespace = cls.namespace
            f.typ = parsed_annotation
    else:
        # Passing a feature in directly is used internally, so not mentioning that in the error message
        raise TypeError(
            (
                f"In order to define feature '{cls.namespace}.{key}', "
                "please set it equal to an underscore expression. "
                f"For example, `{cls.__name__}.{key}: int = _.a + _.b`."
            )
        )
    if f is not None:
        # Process feature field
        f.features_cls = cls
        _process_field(
            f=f,
            error_builder=cls.__chalk_error_builder__,
            comments={},
            class_owner=cls.__chalk_owner__,
            class_tags=tuple(cls.__chalk_tags__),
            class_etl_offline_to_online=cls.__chalk_etl_offline_to_online__,
            class_max_staleness=cls.__chalk_max_staleness__,
        )
        cls.features.append(f)
        wrapped_feature = FeatureWrapper(f)
        type.__setattr__(cls, key, wrapped_feature)
        if is_notebook_defined_feature:
            assert f.unversioned_attribute_name is not None  # for pyright
            FeatureSetBase.__chalk_notebook_defined_feature_fields__[cls.namespace].add(f.unversioned_attribute_name)


def parse_quoted_window_feature(annotation_str: str, module_str: str) -> Type[Windowed]:
    """
    Parses a string like "Windowed[int]", "Windowed[list[int]]", "Windowed[SomeStruct]"
    and returns Windowed with the appropriate type.
    """
    match = re.fullmatch(r"Windowed\[(.+)\]", annotation_str)
    if not match:
        raise ValueError(f"Invalid type format: {annotation_str}")

    inner_type_str = match.group(1)

    try:
        # Handle built-in types and typing constructs
        inner_type = eval(inner_type_str, {"__builtins__": builtins, **typing.__dict__})
    except NameError:
        # If it's not a built-in type, assume it's a user-defined type
        try:
            module = sys.modules.get(module_str, None)
            inner_type = getattr(module, "__dict__", {})[inner_type_str]
        except KeyError:
            raise ValueError(f"Unknown type: {inner_type_str}")

    return Windowed[inner_type]



from __future__ import annotations

import os
from dataclasses import dataclass, field
from typing import Optional

from chalk.utils.environment_parsing import env_var_bool


@dataclass(frozen=True)
class PostgresQueryExecutionParameters:
    attempt_efficient_postgres_execution: bool
    """
    Overrides QueryExecutionParameters.attempt_efficient_parameters if True
    """

    polars_read_csv: bool
    """
    When `attempt_postgres_efficient_execution` is True, this flag decides whether to use polars'
    read_csv or pyarrow's read_csv.
    """

    skip_datetime_timezone_cast: bool
    """
    skip datetime timezone casting, only under efficient execution. This happens BEFORE the sql query.
    """

    csv_read_then_cast: bool = True
    """DEPRECATED"""


@dataclass(frozen=True)
class SnowflakeQueryExecutionParameters:
    snowflake_unload_stage: Optional[str]
    """
    The name of the Snowflake stage to use for unloading data. If None, the default stage will be used.
    I.e. '@CHALK_UNLOAD_STAGE/test_1'
    """


@dataclass(frozen=True)
class BigqueryQueryExecutionParameters:
    bigquery_unload_path: Optional[str]
    """
    The name of the GCS or S3 bucket to use for unloading data. If None, queries will be executed to a temp table.
    """


@dataclass(frozen=True)
class QueryExecutionParameters:
    attempt_efficient_execution: bool
    """
    This will be overriden at query time if the source is a postgres source and
    PostgresQueryExecutionParameters.attempt_efficient_postgres_execution is True in the invoker
    """

    postgres: PostgresQueryExecutionParameters

    snowflake: SnowflakeQueryExecutionParameters

    bigquery: BigqueryQueryExecutionParameters

    yield_empty_batches: bool = False
    """Whether to yield empty batches. This can be useful to capture the schema of an otherwise-empty results set"""

    fallback_to_inefficient_execution: bool = True
    """Whether to fallback to inefficient execution if efficient execution fails for an unexpected error"""

    max_prefetch_size_bytes: int = 1024 * 1024 * 1024
    """If nonnegative, the maximum number of bytes to prefetched when executing a query. If zero or negative,
    then there is no limit to the number of bytes that can be prefetched."""

    num_client_prefetch_threads: int = 4
    """Number of threads to use when downloading query results."""

    tags: frozenset[str] = field(default_factory=frozenset)
    """Tags that can be used to select a source. These tags are proxied from the overlying Chalk query context."""


def query_execution_parameters_from_env_vars():
    """
    For when called in user resolver code.
    If you do not want to do efficient execution, set CHALK_FORCE_SQLALCHEMY_QUERY_EXECUTION_WITHOUT_EXCEPTION to True
    """
    return QueryExecutionParameters(
        attempt_efficient_execution=not env_var_bool("CHALK_FORCE_SQLALCHEMY_QUERY_EXECUTION"),
        max_prefetch_size_bytes=int(os.getenv("CHALK_MAX_PREFETCH_SIZE_BYTES", "1073741824")),
        num_client_prefetch_threads=int(os.getenv("CHALK_NUM_CLIENT_PREFETCH_THREADS", "4")),
        postgres=PostgresQueryExecutionParameters(
            attempt_efficient_postgres_execution=True,
            polars_read_csv=env_var_bool("CHALK_FORCE_POLARS_READ_CSV"),
            skip_datetime_timezone_cast=env_var_bool("CHALK_SKIP_PG_DATETIME_ZONE_CAST"),
        ),
        snowflake=SnowflakeQueryExecutionParameters(
            snowflake_unload_stage=os.getenv("CHALK_SNOWFLAKE_UNLOAD_STAGE", None),
        ),
        bigquery=BigqueryQueryExecutionParameters(
            bigquery_unload_path=os.getenv("CHALK_BIGQUERY_UNLOAD_PATH", None),
        ),
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/deployment.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n chalk/server/v1/deployment.proto\x12\x0f\x63halk.server.v1\x1a\x1fgoogle/protobuf/timestamp.proto"\x88\x01\n\x0eInstanceSizing\x12(\n\rmin_instances\x18\x01 \x01(\rH\x00R\x0cminInstances\x88\x01\x01\x12(\n\rmax_instances\x18\x02 \x01(\rH\x01R\x0cmaxInstances\x88\x01\x01\x42\x10\n\x0e_min_instancesB\x10\n\x0e_max_instances"\xde\x0b\n\nDeployment\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12%\n\x0e\x65nvironment_id\x18\x02 \x01(\tR\renvironmentId\x12\x39\n\x06status\x18\x03 \x01(\x0e\x32!.chalk.server.v1.DeploymentStatusR\x06status\x12\'\n\x0f\x64\x65ployment_tags\x18\x04 \x03(\tR\x0e\x64\x65ploymentTags\x12$\n\x0e\x63loud_build_id\x18\x05 \x01(\tR\x0c\x63loudBuildId\x12!\n\x0ctriggered_by\x18\x06 \x01(\tR\x0btriggeredBy\x12\x38\n\x15requirements_filepath\x18\x07 \x01(\tH\x00R\x14requirementsFilepath\x88\x01\x01\x12\x34\n\x13\x64ockerfile_filepath\x18\x08 \x01(\tH\x01R\x12\x64ockerfileFilepath\x88\x01\x01\x12\x1d\n\x07runtime\x18\t \x01(\tH\x02R\x07runtime\x88\x01\x01\x12\'\n\x0f\x63halkpy_version\x18\n \x01(\tR\x0e\x63halkpyVersion\x12.\n\x13raw_dependency_hash\x18\x0b \x01(\tR\x11rawDependencyHash\x12\x37\n\x15\x66inal_dependency_hash\x18\x0c \x01(\tH\x03R\x13\x66inalDependencyHash\x88\x01\x01\x12\x37\n\x15is_preview_deployment\x18\r \x01(\x08H\x04R\x13isPreviewDeployment\x88\x01\x01\x12\x39\n\ncreated_at\x18\x0e \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x39\n\nupdated_at\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tupdatedAt\x12\x1d\n\ngit_commit\x18\x10 \x01(\tR\tgitCommit\x12\x15\n\x06git_pr\x18\x11 \x01(\tR\x05gitPr\x12\x1d\n\ngit_branch\x18\x12 \x01(\tR\tgitBranch\x12(\n\x10git_author_email\x18\x13 \x01(\tR\x0egitAuthorEmail\x12\x16\n\x06\x62ranch\x18\x14 \x01(\tR\x06\x62ranch\x12)\n\x10project_settings\x18\x15 \x01(\tR\x0fprojectSettings\x12\x32\n\x12requirements_files\x18\x16 \x01(\tH\x05R\x11requirementsFiles\x88\x01\x01\x12\x17\n\x07git_tag\x18\x17 \x01(\tR\x06gitTag\x12$\n\x0e\x62\x61se_image_sha\x18\x18 \x01(\tR\x0c\x62\x61seImageSha\x12\x46\n\x11status_changed_at\x18\x19 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0fstatusChangedAt\x12;\n\x17pinned_platform_version\x18\x1a \x01(\tH\x06R\x15pinnedPlatformVersion\x88\x01\x01\x12\x39\n\x16preview_deployment_tag\x18\x1b \x01(\tH\x07R\x14previewDeploymentTag\x88\x01\x01\x12T\n\x0eprofiling_mode\x18\x1c \x01(\x0e\x32(.chalk.server.v1.DeploymentProfilingModeH\x08R\rprofilingMode\x88\x01\x01\x42\x18\n\x16_requirements_filepathB\x16\n\x14_dockerfile_filepathB\n\n\x08_runtimeB\x18\n\x16_final_dependency_hashB\x18\n\x16_is_preview_deploymentB\x15\n\x13_requirements_filesB\x1a\n\x18_pinned_platform_versionB\x19\n\x17_preview_deployment_tagB\x11\n\x0f_profiling_mode*\xde\x03\n\x10\x44\x65ploymentStatus\x12!\n\x1d\x44\x45PLOYMENT_STATUS_UNSPECIFIED\x10\x00\x12\x1d\n\x19\x44\x45PLOYMENT_STATUS_UNKNOWN\x10\x01\x12\x1d\n\x19\x44\x45PLOYMENT_STATUS_PENDING\x10\x02\x12\x1c\n\x18\x44\x45PLOYMENT_STATUS_QUEUED\x10\x03\x12\x1d\n\x19\x44\x45PLOYMENT_STATUS_WORKING\x10\x04\x12\x1d\n\x19\x44\x45PLOYMENT_STATUS_SUCCESS\x10\x05\x12\x1d\n\x19\x44\x45PLOYMENT_STATUS_FAILURE\x10\x06\x12$\n DEPLOYMENT_STATUS_INTERNAL_ERROR\x10\x07\x12\x1d\n\x19\x44\x45PLOYMENT_STATUS_TIMEOUT\x10\x08\x12\x1f\n\x1b\x44\x45PLOYMENT_STATUS_CANCELLED\x10\t\x12\x1d\n\x19\x44\x45PLOYMENT_STATUS_EXPIRED\x10\n\x12!\n\x1d\x44\x45PLOYMENT_STATUS_BOOT_ERRORS\x10\x0b\x12%\n!DEPLOYMENT_STATUS_AWAITING_SOURCE\x10\x0c\x12\x1f\n\x1b\x44\x45PLOYMENT_STATUS_DEPLOYING\x10\r*\x8a\x01\n\x17\x44\x65ploymentProfilingMode\x12)\n%DEPLOYMENT_PROFILING_MODE_UNSPECIFIED\x10\x00\x12"\n\x1e\x44\x45PLOYMENT_PROFILING_MODE_NONE\x10\x01\x12 \n\x1c\x44\x45PLOYMENT_PROFILING_MODE_O2\x10\x02\x42\x98\x01\n\x13\x63om.chalk.server.v1B\x0f\x44\x65ploymentProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.deployment_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\017DeploymentProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_DEPLOYMENTSTATUS"]._serialized_start = 1731
    _globals["_DEPLOYMENTSTATUS"]._serialized_end = 2209
    _globals["_DEPLOYMENTPROFILINGMODE"]._serialized_start = 2212
    _globals["_DEPLOYMENTPROFILINGMODE"]._serialized_end = 2350
    _globals["_INSTANCESIZING"]._serialized_start = 87
    _globals["_INSTANCESIZING"]._serialized_end = 223
    _globals["_DEPLOYMENT"]._serialized_start = 226
    _globals["_DEPLOYMENT"]._serialized_end = 1728
# @@protoc_insertion_point(module_scope)



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class BloomFilter(_message.Message):
    __slots__ = ("environment", "namespace", "num_entries", "num_expected_entries", "num_hashes", "size_bytes", "data")
    ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    NUM_ENTRIES_FIELD_NUMBER: _ClassVar[int]
    NUM_EXPECTED_ENTRIES_FIELD_NUMBER: _ClassVar[int]
    NUM_HASHES_FIELD_NUMBER: _ClassVar[int]
    SIZE_BYTES_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    environment: str
    namespace: str
    num_entries: int
    num_expected_entries: int
    num_hashes: int
    size_bytes: int
    data: bytes
    def __init__(
        self,
        environment: _Optional[str] = ...,
        namespace: _Optional[str] = ...,
        num_entries: _Optional[int] = ...,
        num_expected_entries: _Optional[int] = ...,
        num_hashes: _Optional[int] = ...,
        size_bytes: _Optional[int] = ...,
        data: _Optional[bytes] = ...,
    ) -> None: ...



from __future__ import annotations

from io import BytesIO
from typing import Any, List, Optional

import pyarrow as pa
import pyarrow.feather as pf

from chalk.client.models import OnlineQueryManyRequest

MAGIC_STR: bytes = "chalk".encode("utf-8")
MULTI_QUERY_MAGIC_STR = "chal1".encode("utf-8")


def write_query_to_buffer(dest: BytesIO, request: OnlineQueryManyRequest, compression: Optional[str] = None) -> None:
    """
    Advances pointer for `dest`
    """
    # called from guarded locations

    data = pa.Table.from_pydict(request.inputs)

    header = request.copy(exclude={"inputs"}).json()
    header_bytes = header.encode("utf-8")
    # Header
    dest.write(len(header_bytes).to_bytes(8, byteorder="big"))
    dest.write(header_bytes)

    # Fill in a placeholder for the body length
    body_length_position = dest.tell()
    dest.write((0).to_bytes(8, byteorder="big"))
    body_start_position = dest.tell()

    # Write the body
    pf.write_feather(data, dest=dest, compression=compression)
    end_of_body = dest.tell()

    # Backfill the body length
    dest.seek(body_length_position)
    dest.write((end_of_body - body_start_position).to_bytes(8, byteorder="big"))

    # Leave the cursor ready to write
    dest.seek(end_of_body)


def decode_multi_query_responses(body: bytes) -> List[Any]:

    response: List[pa.Table] = []

    INT64_BYTE_COUNT = 8

    buffer = BytesIO(body)

    while buffer.tell() < len(body) - 1:
        body_length = int.from_bytes(buffer.read(INT64_BYTE_COUNT), byteorder="big")
        body_buffer = BytesIO(buffer.read(body_length))
        parsed_body = pf.read_table(body_buffer)
        response.append(parsed_body)

    return response



from typing import Any, Optional

import pyarrow as pa

from chalk.operators import StaticExpression


class LiteralExpression(StaticExpression):
    _chalk__expression_name = StaticExpression._chalk__expression_prefix + "literal"

    def __init__(self, value: Any, dtype: pa.DataType, name: Optional[str] = None) -> None:
        super().__init__(pa_value=pa.scalar(value, dtype), literal_name=name)
        self.value = value
        self.dtype = dtype
        self.name = name

    def alias(self, name: str) -> StaticExpression:
        return LiteralExpression(self.value, self.dtype, name)



# pyright: reportPrivateUsage=false
# ^ Underscore objects have chalk-specific fields that are prefixed with _chalk__ to disambiguate from expressions of the form `a.b`
import base64
import dataclasses
from typing import Any, List, Sequence, Union

from chalk._gen.chalk.common.v1.online_query_pb2 import FeatureExpression
from chalk.client.models import OutputExpression
from chalk.features import Feature, FeatureWrapper, Resolver
from chalk.features.feature_set import is_features_cls
from chalk.features.underscore import UnderscoreAttr, UnderscoreCall
from chalk.features.underscore_features import NamedUnderscoreExpr, Underscore, process_named_underscore_expr


@dataclasses.dataclass
class EncodedOutputs:
    string_outputs: List[str]
    feature_expressions_proto: List[FeatureExpression]
    feature_expressions_base64: List[str]  # B64 encoded


@dataclasses.dataclass
class NamespacelessUnderscoreExpr:
    short_name: str
    expr: Underscore


def encode_namespaceless_underscore_proto(expr: NamespacelessUnderscoreExpr) -> FeatureExpression:
    return FeatureExpression(
        namespace="",
        output_column_name=expr.short_name,
        expr=expr.expr._to_proto(),  # pyright: ignore[reportPrivateUsage]
    )


def encode_feature_expression_proto(expr: NamedUnderscoreExpr) -> FeatureExpression:
    processed_expr = process_named_underscore_expr(expr)
    return FeatureExpression(
        namespace=expr.fqn.split(".")[0],
        output_column_name=expr.fqn,
        expr=processed_expr._to_proto(),  # pyright: ignore[reportPrivateUsage]
    )


def encode_feature_expression_base64(expr: FeatureExpression) -> str:
    b = expr.SerializeToString(deterministic=True)
    return base64.b64encode(b).decode("utf-8")


def parse_underscore_aliased_column(u: Underscore) -> tuple[str, Underscore]:
    """
    Given an expression of the form (_.something.etc...).alias("column_name"),
    extracts the underlying expression `_.something.etc....` and the column name.
    i.e. The expected form of `u` is UnderscoreCall(UnderscoreAttr(_.something.etc...., "alias"), "column_name")
    :return: A tuiple of the column name and the underlying expression.
    """
    if not isinstance(u, UnderscoreCall):
        raise ValueError(
            f"Cannot use the given expression '{u}' as a query output -- it must be of the form x.alias('output_column_name')"
        )
    parent = u._chalk__parent
    if not isinstance(parent, UnderscoreAttr) or parent._chalk__attr != "alias":
        raise ValueError(
            f"Cannot use the given expression '{u}' as a query output -- it must be of the form x.alias('output_column_name')"
        )
    original_expr = parent._chalk__parent
    if len(u._chalk__args) != 1:
        raise ValueError(
            f"Cannot use the given expression '{u}' as a query output -- 'alias' method takes one argument, a string column name, e.g. x.alias('output_column_name')"
        )
    column_name = str(u._chalk__args[0])
    return column_name, original_expr


def encode_underscore_alias_proto(column_name: str, expr: Underscore) -> FeatureExpression:
    return FeatureExpression(
        namespace="",
        output_column_name=column_name,
        expr=expr._to_proto(),
    )


def encode_outputs(output: Sequence[Union[str, NamedUnderscoreExpr, Underscore, Any]]) -> EncodedOutputs:
    """Returns a list of encoded outputs and warnings"""
    string_outputs: List[str] = []
    feature_expressions_base64: List[str] = []
    feature_expressions_proto: List[FeatureExpression] = []
    for o in output:
        if isinstance(o, (Feature, FeatureWrapper)):
            string_outputs.append(str(o))
        elif is_features_cls(o):
            string_outputs.append(o.namespace)
        elif isinstance(o, Resolver):
            string_outputs.append(o.fqn.split(".")[-1])
        elif isinstance(o, NamedUnderscoreExpr):
            fe = encode_feature_expression_proto(o)
            feature_expressions_proto.append(fe)
            feature_expressions_base64.append(encode_feature_expression_base64(fe))
        elif isinstance(o, Underscore):
            colname, expr = parse_underscore_aliased_column(o)
            fe = encode_underscore_alias_proto(column_name=colname, expr=expr)
            feature_expressions_proto.append(fe)
            feature_expressions_base64.append(encode_feature_expression_base64(fe))
        else:
            string_outputs.append(str(o))
    return EncodedOutputs(
        string_outputs=string_outputs,
        feature_expressions_base64=feature_expressions_base64,
        feature_expressions_proto=feature_expressions_proto,
    )


def encode_named_underscore(output: NamedUnderscoreExpr) -> OutputExpression:
    fe = encode_feature_expression_proto(output)
    base64_proto = encode_feature_expression_base64(fe)
    return OutputExpression(
        base64_proto=base64_proto,
        python_repr=repr(output),
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/performance_summary.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n)chalk/server/v1/performance_summary.proto\x12\x0f\x63halk.server.v1\x1a\x1cgoogle/protobuf/struct.proto"\xc2\x01\n\x17ShardPerformanceSummary\x12!\n\x0coperation_id\x18\x01 \x01(\tR\x0boperationId\x12\x19\n\x08shard_id\x18\x02 \x01(\x03R\x07shardId\x12i\n%performance_summary_with_query_config\x18\x03 \x01(\x0b\x32\x17.google.protobuf.StructR!performanceSummaryWithQueryConfig"V\n2ListOfflineQueryShardPerformanceSummariesPageToken\x12 \n\x0cshard_id_hwm\x18\x01 \x01(\x03R\nshardIdHwm"\x8a\x01\n0ListOfflineQueryShardPerformanceSummariesRequest\x12!\n\x0coperation_id\x18\x01 \x01(\tR\x0boperationId\x12\x14\n\x05limit\x18\x02 \x01(\x03R\x05limit\x12\x1d\n\npage_token\x18\x03 \x01(\tR\tpageToken"\xba\x01\n1ListOfflineQueryShardPerformanceSummariesResponse\x12]\n\x15performance_summaries\x18\x01 \x03(\x0b\x32(.chalk.server.v1.ShardPerformanceSummaryR\x14performanceSummaries\x12&\n\x0fnext_page_token\x18\x02 \x01(\tR\rnextPageTokenB\xa0\x01\n\x13\x63om.chalk.server.v1B\x17PerformanceSummaryProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.performance_summary_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\027PerformanceSummaryProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_SHARDPERFORMANCESUMMARY"]._serialized_start = 93
    _globals["_SHARDPERFORMANCESUMMARY"]._serialized_end = 287
    _globals["_LISTOFFLINEQUERYSHARDPERFORMANCESUMMARIESPAGETOKEN"]._serialized_start = 289
    _globals["_LISTOFFLINEQUERYSHARDPERFORMANCESUMMARIESPAGETOKEN"]._serialized_end = 375
    _globals["_LISTOFFLINEQUERYSHARDPERFORMANCESUMMARIESREQUEST"]._serialized_start = 378
    _globals["_LISTOFFLINEQUERYSHARDPERFORMANCESUMMARIESREQUEST"]._serialized_end = 516
    _globals["_LISTOFFLINEQUERYSHARDPERFORMANCESUMMARIESRESPONSE"]._serialized_start = 519
    _globals["_LISTOFFLINEQUERYSHARDPERFORMANCESUMMARIESRESPONSE"]._serialized_end = 705
# @@protoc_insertion_point(module_scope)



from inspect import findsource
from types import FrameType
from typing import Tuple


def get_source_object_starting(f: FrameType) -> Tuple[str, int, int]:
    start = f.f_lineno - 1
    lines, _ = findsource(f)
    start = max(start, 1)
    source_code = ""
    for i, cc in enumerate(lines[start:]):
        source_code += cc
        if cc.strip().endswith(")"):
            break
    # +1 to correct for 0 indexing
    return source_code, start + 1, start + i + 1



from __future__ import annotations

import uuid
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Union

from typing_extensions import Self

from chalk.client import ChalkError

if TYPE_CHECKING:
    from pydantic import BaseModel
else:
    try:
        from pydantic.v1 import BaseModel
    except ImportError:
        from pydantic import BaseModel


class BatchProgress(BaseModel):
    total: int = 0
    computed: int = 0
    failed: int = 0
    start: Optional[datetime]
    end: Optional[datetime]
    total_duration_s: float = 0.0

    @classmethod
    def empty(cls) -> "BatchProgress":
        return BatchProgress(
            start=None,
            end=None,
            total=0,
            computed=0,
            failed=0,
            total_duration_s=0.0,
        )

    def __add__(self, other: BatchProgress) -> Self:
        if not isinstance(other, BatchProgress):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise NotImplementedError(
                f"Can only add ProgressReport to ProgressReport. Received '{type(other).__name__}'"
            )

        return type(self)(
            total=self.total + other.total,
            computed=self.computed + other.computed,
            failed=self.failed + other.failed,
            start=(
                self.start
                if other.start is None
                else (other.start if self.start is None else min(self.start, other.start))
            ),
            total_duration_s=self.total_duration_s + other.total_duration_s,
            end=None,
        )


class BatchProgressSum(BaseModel):
    total: int = 0
    computed: int = 0
    failed: int = 0
    total_duration_s: float = 0.0

    @classmethod
    def from_progresses(cls, *args: Union[BatchProgress, Iterable[BatchProgress]]) -> "BatchProgressSum":
        summed = BatchProgressSum()
        for arg in args:
            if not isinstance(arg, BatchProgress):
                for a in arg:
                    summed += a
            else:
                summed += arg

        return summed

    def __add__(self, other: BatchProgress) -> Self:
        if not isinstance(other, BatchProgress):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise NotImplementedError(f"Can only add BatchProgress to BatchProgress. Received '{type(other).__name__}'")

        return type(self)(
            total=self.total + other.total,
            computed=self.computed + other.computed,
            failed=self.failed + other.failed,
            total_duration_s=self.total_duration_s + other.total_duration_s,
        )


class BatchOpKind(str, Enum):
    OFFLINE_QUERY = "OFFLINE_QUERY"
    RECOMPUTE = "RECOMPUTE"
    CRON = "CRON"
    AGGREGATION_BACKFILL = "AGGREGATION_BACKFILL"


class BatchOpStatus(str, Enum):
    INIT = "INIT"
    COMPUTE_STARTED = "COMPUTE_STARTED"
    COMPUTE_ENDED = "COMPUTE_ENDED"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


class ChunkReport(BaseModel):
    # doesn't have its own status.
    # resolver encapsulates the
    # status of the chunk.
    progress: BatchProgress
    generated_at: datetime


class BatchResolverReport(BaseModel):
    resolver_fqn: str
    status: BatchOpStatus
    chunks: List[ChunkReport]
    progress: BatchProgress
    generated_at: datetime
    error: Optional[ChalkError]
    all_errors: Optional[List[ChalkError]] = None


class BatchReport(BaseModel):
    operation_id: str
    operation_kind: BatchOpKind
    status: BatchOpStatus
    resolvers: List[BatchResolverReport]
    progress: BatchProgress
    environment_id: str
    team_id: str
    deployment_id: str
    error: Optional[ChalkError]
    generated_at: datetime
    all_errors: Optional[List[ChalkError]] = None
    operation_metadata: Optional[Dict[str, str]] = None
    computer_id: Optional[int] = None


class BatchReportResponse(BaseModel):
    report: Optional[BatchReport]
    error: Optional[ChalkError] = None


class InitiateOfflineQueryResponse(BaseModel):
    revision_id: uuid.UUID


class Heartbeat(BaseModel):
    operation_id: str
    operation_kind: BatchOpKind
    project_id: str
    environment_id: str
    team_id: str
    deployment_id: str
    generated_at: datetime
    computer_id: int
    shard_id: int



from chalk._gen.chalk.chart.v1 import densetimeserieschart_pb2 as _densetimeserieschart_pb2
from chalk._gen.chalk.common.v1 import chart_pb2 as _chart_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class FeatureValueAggregation(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FEATURE_VALUE_AGGREGATION_UNSPECIFIED: _ClassVar[FeatureValueAggregation]
    FEATURE_VALUE_AGGREGATION_UNIQUE_VALUES: _ClassVar[FeatureValueAggregation]
    FEATURE_VALUE_AGGREGATION_TOTAL_OBSERVATIONS: _ClassVar[FeatureValueAggregation]
    FEATURE_VALUE_AGGREGATION_NULL_PERCENTAGE: _ClassVar[FeatureValueAggregation]
    FEATURE_VALUE_AGGREGATION_MAX_VALUE: _ClassVar[FeatureValueAggregation]
    FEATURE_VALUE_AGGREGATION_MIN_VALUE: _ClassVar[FeatureValueAggregation]
    FEATURE_VALUE_AGGREGATION_AVERAGE: _ClassVar[FeatureValueAggregation]
    FEATURE_VALUE_AGGREGATION_UNIQUE_PKEYS: _ClassVar[FeatureValueAggregation]
    FEATURE_VALUE_AGGREGATION_P95: _ClassVar[FeatureValueAggregation]
    FEATURE_VALUE_AGGREGATION_P75: _ClassVar[FeatureValueAggregation]
    FEATURE_VALUE_AGGREGATION_P50: _ClassVar[FeatureValueAggregation]
    FEATURE_VALUE_AGGREGATION_P25: _ClassVar[FeatureValueAggregation]
    FEATURE_VALUE_AGGREGATION_P05: _ClassVar[FeatureValueAggregation]

FEATURE_VALUE_AGGREGATION_UNSPECIFIED: FeatureValueAggregation
FEATURE_VALUE_AGGREGATION_UNIQUE_VALUES: FeatureValueAggregation
FEATURE_VALUE_AGGREGATION_TOTAL_OBSERVATIONS: FeatureValueAggregation
FEATURE_VALUE_AGGREGATION_NULL_PERCENTAGE: FeatureValueAggregation
FEATURE_VALUE_AGGREGATION_MAX_VALUE: FeatureValueAggregation
FEATURE_VALUE_AGGREGATION_MIN_VALUE: FeatureValueAggregation
FEATURE_VALUE_AGGREGATION_AVERAGE: FeatureValueAggregation
FEATURE_VALUE_AGGREGATION_UNIQUE_PKEYS: FeatureValueAggregation
FEATURE_VALUE_AGGREGATION_P95: FeatureValueAggregation
FEATURE_VALUE_AGGREGATION_P75: FeatureValueAggregation
FEATURE_VALUE_AGGREGATION_P50: FeatureValueAggregation
FEATURE_VALUE_AGGREGATION_P25: FeatureValueAggregation
FEATURE_VALUE_AGGREGATION_P05: FeatureValueAggregation

class GetFeatureValuesChartRequest(_message.Message):
    __slots__ = ("fqn", "aggregate_by", "window_period", "start_ms", "end_ms")
    FQN_FIELD_NUMBER: _ClassVar[int]
    AGGREGATE_BY_FIELD_NUMBER: _ClassVar[int]
    WINDOW_PERIOD_FIELD_NUMBER: _ClassVar[int]
    START_MS_FIELD_NUMBER: _ClassVar[int]
    END_MS_FIELD_NUMBER: _ClassVar[int]
    fqn: str
    aggregate_by: _containers.RepeatedScalarFieldContainer[FeatureValueAggregation]
    window_period: str
    start_ms: int
    end_ms: int
    def __init__(
        self,
        fqn: _Optional[str] = ...,
        aggregate_by: _Optional[_Iterable[_Union[FeatureValueAggregation, str]]] = ...,
        window_period: _Optional[str] = ...,
        start_ms: _Optional[int] = ...,
        end_ms: _Optional[int] = ...,
    ) -> None: ...

class GetFeatureValuesChartResponse(_message.Message):
    __slots__ = ("chart",)
    CHART_FIELD_NUMBER: _ClassVar[int]
    chart: _chart_pb2.Chart
    def __init__(self, chart: _Optional[_Union[_chart_pb2.Chart, _Mapping]] = ...) -> None: ...

class GetFeatureValuesTimeSeriesChartRequest(_message.Message):
    __slots__ = ("fqn", "aggregate_by", "window_period", "start_timestamp_inclusive", "end_timestamp_exclusive")
    FQN_FIELD_NUMBER: _ClassVar[int]
    AGGREGATE_BY_FIELD_NUMBER: _ClassVar[int]
    WINDOW_PERIOD_FIELD_NUMBER: _ClassVar[int]
    START_TIMESTAMP_INCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    END_TIMESTAMP_EXCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    fqn: str
    aggregate_by: _containers.RepeatedScalarFieldContainer[FeatureValueAggregation]
    window_period: _duration_pb2.Duration
    start_timestamp_inclusive: _timestamp_pb2.Timestamp
    end_timestamp_exclusive: _timestamp_pb2.Timestamp
    def __init__(
        self,
        fqn: _Optional[str] = ...,
        aggregate_by: _Optional[_Iterable[_Union[FeatureValueAggregation, str]]] = ...,
        window_period: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
        start_timestamp_inclusive: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        end_timestamp_exclusive: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
    ) -> None: ...

class GetFeatureValuesTimeSeriesChartResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetFeatureValuesTimeSeriesChartV2Request(_message.Message):
    __slots__ = ("fqn", "aggregate_by", "window_period", "start_timestamp_inclusive", "end_timestamp_exclusive")
    FQN_FIELD_NUMBER: _ClassVar[int]
    AGGREGATE_BY_FIELD_NUMBER: _ClassVar[int]
    WINDOW_PERIOD_FIELD_NUMBER: _ClassVar[int]
    START_TIMESTAMP_INCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    END_TIMESTAMP_EXCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    fqn: str
    aggregate_by: _containers.RepeatedScalarFieldContainer[FeatureValueAggregation]
    window_period: _duration_pb2.Duration
    start_timestamp_inclusive: _timestamp_pb2.Timestamp
    end_timestamp_exclusive: _timestamp_pb2.Timestamp
    def __init__(
        self,
        fqn: _Optional[str] = ...,
        aggregate_by: _Optional[_Iterable[_Union[FeatureValueAggregation, str]]] = ...,
        window_period: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
        start_timestamp_inclusive: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        end_timestamp_exclusive: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
    ) -> None: ...

class GetFeatureValuesTimeSeriesChartV2Response(_message.Message):
    __slots__ = ("chart",)
    CHART_FIELD_NUMBER: _ClassVar[int]
    chart: _densetimeserieschart_pb2.DenseTimeSeriesChart
    def __init__(
        self, chart: _Optional[_Union[_densetimeserieschart_pb2.DenseTimeSeriesChart, _Mapping]] = ...
    ) -> None: ...



import itertools
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List, Literal, Mapping, Optional, cast

import yaml

from chalk._version import __version__
from chalk.config._validator import Validator


@dataclass
class EnvironmentSettings:
    runtime: Optional[Literal["python310", "python311", "python312"]]
    requirements: Optional[str]
    dockerfile: Optional[str]
    platform_version: Optional[str]

    @staticmethod
    def from_py(value: Any, environment_name: str) -> "EnvironmentSettings":
        prefix = f"environments.{environment_name}"
        value = Validator.dict_with_str_keys(value, name=prefix)
        return EnvironmentSettings(
            runtime=cast(
                Optional[Literal["python310", "python311", "python312"]],
                Validator.optional_string(
                    value.get("runtime"),
                    f"{prefix}.runtime",
                    one_of=("python310", "python311", "python312"),
                ),
            ),
            requirements=Validator.optional_string(value.get("requirements"), f"{prefix}.requirements"),
            dockerfile=Validator.optional_string(value.get("dockerfile"), f"{prefix}.dockerfile"),
            platform_version=Validator.optional_string(value.get("platform_version"), f"{prefix}.platform_version"),
        )

    def as_dict(self):
        return {
            "runtime": self.runtime,
            "requirements": self.requirements,
            "dockerfile": self.dockerfile,
            "platform_version": self.platform_version,
        }


@dataclass
class MetadataSettings:
    name: str
    missing: str

    @staticmethod
    def from_py(value: Any, kind: str) -> "MetadataSettings":
        value = Validator.dict_with_str_keys(value, name=f"validation.{kind}.metadata")
        return MetadataSettings(
            name=Validator.string(value.get("name"), f"validation.{kind}.metadata.name"),
            missing=Validator.string(value.get("missing"), f"validation.{kind}.metadata.missing"),
        )

    def as_dict(self):
        return {
            "name": self.name,
            "missing": self.missing,
        }


@dataclass
class FeatureSettings:
    metadata: Optional[List[MetadataSettings]]

    @staticmethod
    def from_py(value: Any) -> "FeatureSettings":
        value = Validator.dict_with_str_keys(value, name="validation.feature")
        metadata = value.get("metadata", None)
        if metadata is None:
            return FeatureSettings(metadata=None)
        if not isinstance(metadata, list):
            raise ValueError(f"Expected list, got '{metadata}' for validation.feature.metadata")
        return FeatureSettings(metadata=[MetadataSettings.from_py(m, "feature") for m in metadata])

    def as_dict(self):
        return {
            "metadata": [m.as_dict() for m in self.metadata] if self.metadata is not None else None,
        }


@dataclass
class ResolverSettings:
    metadata: Optional[List[MetadataSettings]]

    @staticmethod
    def from_py(value: Any) -> "ResolverSettings":
        value = Validator.dict_with_str_keys(value, name="validation.resolver")
        metadata = value.get("metadata", None)
        if metadata is None:
            return ResolverSettings(metadata=None)
        if not isinstance(metadata, list):
            raise ValueError(f"Expected list, got '{metadata}' for validation.resolver.metadata")
        return ResolverSettings(metadata=[MetadataSettings.from_py(m, "resolver") for m in metadata])

    def as_dict(self):
        return {
            "metadata": [m.as_dict() for m in self.metadata] if self.metadata is not None else None,
        }


@dataclass
class ValidationSettings:
    feature: Optional[FeatureSettings]
    resolver: Optional[ResolverSettings]

    @staticmethod
    def from_py(value: Any) -> "ValidationSettings":
        d = Validator.dict_with_str_keys(value, name="validation")
        raw_feature = d.get("feature", None)
        raw_resolver = d.get("resolver", None)
        return ValidationSettings(
            feature=FeatureSettings.from_py(raw_feature) if raw_feature is not None else None,
            resolver=ResolverSettings.from_py(raw_resolver) if raw_resolver is not None else None,
        )

    def as_dict(self):
        return {
            "feature": self.feature.as_dict() if self.feature is not None else None,
            "resolver": self.resolver.as_dict() if self.resolver is not None else None,
        }


@dataclass
class ProjectSettings:
    project: str
    environments: Optional[Mapping[str, EnvironmentSettings]]
    validation: Optional[ValidationSettings]
    local_path: str
    chalkpy: str

    @staticmethod
    def from_py(value: Any, local_path: str):
        value = Validator.dict_with_str_keys(value, name="root")
        raw_envs = Validator.dict_with_str_keys_or_none(value.get("environments", None), "environments")
        raw_validation = value.get("validation", None)
        return ProjectSettings(
            project=Validator.string(value.get("project"), "project"),
            environments=(
                None
                if raw_envs is None
                else {
                    environment_name: EnvironmentSettings.from_py(
                        value=v,
                        environment_name=environment_name,
                    )
                    for environment_name, v in raw_envs.items()
                }
            ),
            validation=ValidationSettings.from_py(raw_validation) if raw_validation is not None else None,
            local_path=local_path,
            chalkpy=__version__,
        )

    def as_dict(self):
        return {
            "project": self.project,
            "environments": {k: v.as_dict() for k, v in self.environments.items()} if self.environments else None,
            "validation": self.validation.as_dict() if self.validation else None,
            "local_path": self.local_path,
            "chalkpy": self.chalkpy,
        }


def _load_project_config_at_path(filename: Path) -> Optional[ProjectSettings]:
    has_default_requirements = os.path.exists(filename.parent / "requirements.txt")
    try:
        with open(filename, "r") as f:
            parsed = yaml.safe_load(f)
            settings = ProjectSettings.from_py(
                parsed,
                local_path=str(filename.absolute().resolve()),
            )
            if has_default_requirements and settings.environments is not None:
                for cfg in settings.environments.values():
                    if cfg.requirements is None:
                        cfg.requirements = "requirements.txt"
            return settings
    except OSError:
        return None
    except ValueError as e:
        raise ValueError(f"Failed to load project config (chalkpy=={__version__}): {e}")


def load_project_config() -> Optional[ProjectSettings]:
    base = Path(os.getcwd())

    for d in itertools.chain([base], base.parents):
        project = _load_project_config_at_path(d / "chalk.yaml") or _load_project_config_at_path(d / "chalk.yml")

        if project is not None:
            return project

    return None



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.server.v1 import scheduled_query_run_pb2 as _scheduled_query_run_pb2
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.kube_pb2 import (
    GetKubernetesEventsRequest,
    GetKubernetesEventsResponse,
    GetKubernetesPersistentVolumesRequest,
    GetKubernetesPersistentVolumesResponse,
    GetPodStackTraceDumpRequest,
    GetPodStackTraceDumpResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class KubeServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    GetPodStackTraceDump: UnaryUnaryMultiCallable[
        GetPodStackTraceDumpRequest,
        GetPodStackTraceDumpResponse,
    ]
    """GetPodStackTraceDump gets the stack trace dump from a single process running in a pod
    The process can be specified either by name or process ID
    """
    GetKubernetesEvents: UnaryUnaryMultiCallable[
        GetKubernetesEventsRequest,
        GetKubernetesEventsResponse,
    ]
    GetKubernetesPersistentVolumes: UnaryUnaryMultiCallable[
        GetKubernetesPersistentVolumesRequest,
        GetKubernetesPersistentVolumesResponse,
    ]

class KubeServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def GetPodStackTraceDump(
        self,
        request: GetPodStackTraceDumpRequest,
        context: ServicerContext,
    ) -> GetPodStackTraceDumpResponse:
        """GetPodStackTraceDump gets the stack trace dump from a single process running in a pod
        The process can be specified either by name or process ID
        """
    @abstractmethod
    def GetKubernetesEvents(
        self,
        request: GetKubernetesEventsRequest,
        context: ServicerContext,
    ) -> GetKubernetesEventsResponse: ...
    @abstractmethod
    def GetKubernetesPersistentVolumes(
        self,
        request: GetKubernetesPersistentVolumesRequest,
        context: ServicerContext,
    ) -> GetKubernetesPersistentVolumesResponse: ...

def add_KubeServiceServicer_to_server(servicer: KubeServiceServicer, server: Server) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.cli_pb2 import (
    GetVersionsRequest,
    GetVersionsResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class CommandLineInterfaceServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    GetVersions: UnaryUnaryMultiCallable[
        GetVersionsRequest,
        GetVersionsResponse,
    ]

class CommandLineInterfaceServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def GetVersions(
        self,
        request: GetVersionsRequest,
        context: ServicerContext,
    ) -> GetVersionsResponse: ...

def add_CommandLineInterfaceServiceServicer_to_server(
    servicer: CommandLineInterfaceServiceServicer, server: Server
) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/http_engine_proxy.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from chalk._gen.chalk.common.v1 import chalk_error_pb2 as chalk_dot_common_dot_v1_dot_chalk__error__pb2
from chalk._gen.chalk.common.v1 import dataset_response_pb2 as chalk_dot_common_dot_v1_dot_dataset__response__pb2
from chalk._gen.chalk.common.v1 import offline_query_pb2 as chalk_dot_common_dot_v1_dot_offline__query__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\'chalk/server/v1/http_engine_proxy.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a!chalk/common/v1/chalk_error.proto\x1a&chalk/common/v1/dataset_response.proto\x1a#chalk/common/v1/offline_query.proto"[\n\x19\x43reateOfflineQueryRequest\x12>\n\x07request\x18\x01 \x01(\x0b\x32$.chalk.common.v1.OfflineQueryRequestR\x07request"\xd1\x02\n\x1a\x43reateOfflineQueryResponse\x12#\n\x0bis_finished\x18\x01 \x01(\x08\x42\x02\x18\x01R\nisFinished\x12\x1c\n\x07version\x18\x02 \x01(\x05\x42\x02\x18\x01R\x07version\x12\x16\n\x04urls\x18\x03 \x03(\tB\x02\x18\x01R\x04urls\x12\x37\n\x06\x65rrors\x18\x04 \x03(\x0b\x32\x1b.chalk.common.v1.ChalkErrorB\x02\x18\x01R\x06\x65rrors\x12\x46\n\x07\x63olumns\x18\x05 \x01(\x0b\x32#.chalk.common.v1.ColumnMetadataListB\x02\x18\x01H\x00R\x07\x63olumns\x88\x01\x01\x12K\n\x10\x64\x61taset_response\x18\x06 \x01(\x0b\x32 .chalk.common.v1.DatasetResponseR\x0f\x64\x61tasetResponseB\n\n\x08_columns2\x8c\x01\n\x16HTTPEngineProxyService\x12r\n\x12\x43reateOfflineQuery\x12*.chalk.server.v1.CreateOfflineQueryRequest\x1a+.chalk.server.v1.CreateOfflineQueryResponse"\x03\x80}\x04\x42\x9d\x01\n\x13\x63om.chalk.server.v1B\x14HttpEngineProxyProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.http_engine_proxy_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\024HttpEngineProxyProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_CREATEOFFLINEQUERYRESPONSE"].fields_by_name["is_finished"]._options = None
    _globals["_CREATEOFFLINEQUERYRESPONSE"].fields_by_name["is_finished"]._serialized_options = b"\030\001"
    _globals["_CREATEOFFLINEQUERYRESPONSE"].fields_by_name["version"]._options = None
    _globals["_CREATEOFFLINEQUERYRESPONSE"].fields_by_name["version"]._serialized_options = b"\030\001"
    _globals["_CREATEOFFLINEQUERYRESPONSE"].fields_by_name["urls"]._options = None
    _globals["_CREATEOFFLINEQUERYRESPONSE"].fields_by_name["urls"]._serialized_options = b"\030\001"
    _globals["_CREATEOFFLINEQUERYRESPONSE"].fields_by_name["errors"]._options = None
    _globals["_CREATEOFFLINEQUERYRESPONSE"].fields_by_name["errors"]._serialized_options = b"\030\001"
    _globals["_CREATEOFFLINEQUERYRESPONSE"].fields_by_name["columns"]._options = None
    _globals["_CREATEOFFLINEQUERYRESPONSE"].fields_by_name["columns"]._serialized_options = b"\030\001"
    _globals["_HTTPENGINEPROXYSERVICE"].methods_by_name["CreateOfflineQuery"]._options = None
    _globals["_HTTPENGINEPROXYSERVICE"].methods_by_name["CreateOfflineQuery"]._serialized_options = b"\200}\004"
    _globals["_CREATEOFFLINEQUERYREQUEST"]._serialized_start = 205
    _globals["_CREATEOFFLINEQUERYREQUEST"]._serialized_end = 296
    _globals["_CREATEOFFLINEQUERYRESPONSE"]._serialized_start = 299
    _globals["_CREATEOFFLINEQUERYRESPONSE"]._serialized_end = 636
    _globals["_HTTPENGINEPROXYSERVICE"]._serialized_start = 639
    _globals["_HTTPENGINEPROXYSERVICE"]._serialized_end = 779
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.status_pb2 import (
    CheckHealthRequest,
    CheckHealthResponse,
    GetClusterMetricsRequest,
    GetClusterMetricsResponse,
    GetHealthRequest,
    GetHealthResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class HealthServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    CheckHealth: UnaryUnaryMultiCallable[
        CheckHealthRequest,
        CheckHealthResponse,
    ]
    """If any checks fail, this request fails."""
    GetHealth: UnaryUnaryMultiCallable[
        GetHealthRequest,
        GetHealthResponse,
    ]
    """Return the actual health check values."""
    GetClusterMetrics: UnaryUnaryMultiCallable[
        GetClusterMetricsRequest,
        GetClusterMetricsResponse,
    ]
    """Return collected cluster prometheus metrics"""

class HealthServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def CheckHealth(
        self,
        request: CheckHealthRequest,
        context: ServicerContext,
    ) -> CheckHealthResponse:
        """If any checks fail, this request fails."""
    @abstractmethod
    def GetHealth(
        self,
        request: GetHealthRequest,
        context: ServicerContext,
    ) -> GetHealthResponse:
        """Return the actual health check values."""
    @abstractmethod
    def GetClusterMetrics(
        self,
        request: GetClusterMetricsRequest,
        context: ServicerContext,
    ) -> GetClusterMetricsResponse:
        """Return collected cluster prometheus metrics"""

def add_HealthServiceServicer_to_server(servicer: HealthServiceServicer, server: Server) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.scheduled_query_run_pb2 import (
    GetScheduledQueryRunRequest,
    GetScheduledQueryRunResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class ScheduledQueryServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    GetScheduledQueryRun: UnaryUnaryMultiCallable[
        GetScheduledQueryRunRequest,
        GetScheduledQueryRunResponse,
    ]

class ScheduledQueryServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def GetScheduledQueryRun(
        self,
        request: GetScheduledQueryRunRequest,
        context: ServicerContext,
    ) -> GetScheduledQueryRunResponse: ...

def add_ScheduledQueryServiceServicer_to_server(servicer: ScheduledQueryServiceServicer, server: Server) -> None: ...



import dataclasses
from typing import Generic, Mapping, Optional, Type, TypeVar, get_args

import pyarrow as pa

T = TypeVar("T", str, bytes)


@dataclasses.dataclass
class HttpResponse(Generic[T]):
    """
     Feature annotation for HTTP responses, with a string body. HTTP responses are treated internally as
     structs, and you can treat them as such in code/expressions that involve this feature.

     The underlying pyarrow type of HttpStringResponse is
     pa.struct_([
         pa.field("status_code", pa.int64()),
         pa.field("headers", pa.map_(pa.large_string(), pa.large_string())),
         pa.field("body", pa.large_string()),
         pa.field("final_url", pa.large_string()),
     ])

    Examples
     --------
     >>> from chalk import _
     >>> import chalk.functions as F
     >>> from chalk.functions.http import HttpResponse
     >>> @features
     ... class User:
     ...    id: str
     ...    resp: HttpResponse[str] = F.http_request("https://example.com", "GET")
     ...    status_code: int = _.resp.status_code
     ...    resp_body: str = _.resp.body
    """

    status_code: Optional[int]
    headers: Optional[Mapping[str, str]]
    body: Optional[T]
    final_url: Optional[str]


def get_http_response_as_pyarrow(t: Type[HttpResponse]) -> pa.DataType:
    type_args = get_args(t)
    if len(type_args) != 1:
        raise TypeError(f"Unsupported http-response type: {t}")
    if issubclass(type_args[0], str):
        return pa.struct(
            [
                pa.field("status_code", pa.int64()),
                pa.field("headers", pa.map_(pa.large_utf8(), pa.large_utf8())),
                pa.field("body", pa.large_utf8()),
                pa.field("final_url", pa.large_utf8()),
            ]
        )
    if issubclass(type_args[0], bytes):
        return pa.struct(
            [
                pa.field("status_code", pa.int64()),
                pa.field("headers", pa.map_(pa.large_utf8(), pa.large_utf8())),
                pa.field("body", pa.large_binary()),
                pa.field("final_url", pa.large_utf8()),
            ]
        )
    raise TypeError(f"Unsupported http-response type: {t}")


__all__ = ["HttpResponse", "get_http_response_as_pyarrow"]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/auth/v1/audit.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x19\x63halk/auth/v1/audit.proto\x12\rchalk.auth.v1\x1a google/protobuf/descriptor.proto"a\n\x0c\x41uditOptions\x12/\n\x05level\x18\x01 \x01(\x0e\x32\x19.chalk.auth.v1.AuditLevelR\x05level\x12 \n\x0b\x64\x65scription\x18\x02 \x01(\tR\x0b\x64\x65scription*V\n\nAuditLevel\x12\x1b\n\x17\x41UDIT_LEVEL_UNSPECIFIED\x10\x00\x12\x16\n\x12\x41UDIT_LEVEL_ERRORS\x10\x01\x12\x13\n\x0f\x41UDIT_LEVEL_ALL\x10\x02:V\n\x05\x61udit\x12\x1e.google.protobuf.MethodOptions\x18\xb1\xea\x01 \x01(\x0b\x32\x1b.chalk.auth.v1.AuditOptionsR\x05\x61udit\x88\x01\x01\x42\x85\x01\n\x11\x63om.chalk.auth.v1B\nAuditProtoP\x01Z\x0e\x61uth/v1;authv1\xa2\x02\x03\x43\x41X\xaa\x02\rChalk.Auth.V1\xca\x02\rChalk\\Auth\\V1\xe2\x02\x19\x43halk\\Auth\\V1\\GPBMetadata\xea\x02\x0f\x43halk::Auth::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.auth.v1.audit_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\021com.chalk.auth.v1B\nAuditProtoP\001Z\016auth/v1;authv1\242\002\003CAX\252\002\rChalk.Auth.V1\312\002\rChalk\\Auth\\V1\342\002\031Chalk\\Auth\\V1\\GPBMetadata\352\002\017Chalk::Auth::V1"
    _globals["_AUDITLEVEL"]._serialized_start = 177
    _globals["_AUDITLEVEL"]._serialized_end = 263
    _globals["_AUDITOPTIONS"]._serialized_start = 78
    _globals["_AUDITOPTIONS"]._serialized_end = 175
# @@protoc_insertion_point(module_scope)



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class HealthCheckStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    HEALTH_CHECK_STATUS_UNSPECIFIED: _ClassVar[HealthCheckStatus]
    HEALTH_CHECK_STATUS_OK: _ClassVar[HealthCheckStatus]
    HEALTH_CHECK_STATUS_FAILING: _ClassVar[HealthCheckStatus]

HEALTH_CHECK_STATUS_UNSPECIFIED: HealthCheckStatus
HEALTH_CHECK_STATUS_OK: HealthCheckStatus
HEALTH_CHECK_STATUS_FAILING: HealthCheckStatus

class HealthCheck(_message.Message):
    __slots__ = ("name", "status", "message", "latency", "kube_data", "metadata")
    class MetadataEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    NAME_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    LATENCY_FIELD_NUMBER: _ClassVar[int]
    KUBE_DATA_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    name: str
    status: HealthCheckStatus
    message: str
    latency: _duration_pb2.Duration
    kube_data: _struct_pb2.Struct
    metadata: _containers.ScalarMap[str, str]
    def __init__(
        self,
        name: _Optional[str] = ...,
        status: _Optional[_Union[HealthCheckStatus, str]] = ...,
        message: _Optional[str] = ...,
        latency: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
        kube_data: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...,
        metadata: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class HealthCheckFilters(_message.Message):
    __slots__ = ("name", "status")
    NAME_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    name: _containers.RepeatedScalarFieldContainer[str]
    status: _containers.RepeatedScalarFieldContainer[HealthCheckStatus]
    def __init__(
        self, name: _Optional[_Iterable[str]] = ..., status: _Optional[_Iterable[_Union[HealthCheckStatus, str]]] = ...
    ) -> None: ...

class CheckHealthRequest(_message.Message):
    __slots__ = ("filters",)
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    filters: HealthCheckFilters
    def __init__(self, filters: _Optional[_Union[HealthCheckFilters, _Mapping]] = ...) -> None: ...

class CheckHealthResponse(_message.Message):
    __slots__ = ("checks",)
    CHECKS_FIELD_NUMBER: _ClassVar[int]
    checks: _containers.RepeatedCompositeFieldContainer[HealthCheck]
    def __init__(self, checks: _Optional[_Iterable[_Union[HealthCheck, _Mapping]]] = ...) -> None: ...

class GetHealthRequest(_message.Message):
    __slots__ = ("filters",)
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    filters: HealthCheckFilters
    def __init__(self, filters: _Optional[_Union[HealthCheckFilters, _Mapping]] = ...) -> None: ...

class GetHealthResponse(_message.Message):
    __slots__ = ("checks",)
    CHECKS_FIELD_NUMBER: _ClassVar[int]
    checks: _containers.RepeatedCompositeFieldContainer[HealthCheck]
    def __init__(self, checks: _Optional[_Iterable[_Union[HealthCheck, _Mapping]]] = ...) -> None: ...

class GetClusterMetricsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetClusterMetricsResponse(_message.Message):
    __slots__ = ("metrics",)
    METRICS_FIELD_NUMBER: _ClassVar[int]
    metrics: str
    def __init__(self, metrics: _Optional[str] = ...) -> None: ...



from chalk._gen.chalk.auth.v1 import audit_pb2 as _audit_pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.utils.v1 import sensitive_pb2 as _sensitive_pb2
from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class PagerDutySeverity(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    PAGER_DUTY_SEVERITY_UNSPECIFIED: _ClassVar[PagerDutySeverity]
    PAGER_DUTY_SEVERITY_INFO: _ClassVar[PagerDutySeverity]
    PAGER_DUTY_SEVERITY_WARNING: _ClassVar[PagerDutySeverity]
    PAGER_DUTY_SEVERITY_ERROR: _ClassVar[PagerDutySeverity]
    PAGER_DUTY_SEVERITY_CRITICAL: _ClassVar[PagerDutySeverity]

class PagerDutyEventAction(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    PAGER_DUTY_EVENT_ACTION_UNSPECIFIED: _ClassVar[PagerDutyEventAction]
    PAGER_DUTY_EVENT_ACTION_TRIGGER: _ClassVar[PagerDutyEventAction]
    PAGER_DUTY_EVENT_ACTION_ACKNOWLEDGE: _ClassVar[PagerDutyEventAction]
    PAGER_DUTY_EVENT_ACTION_RESOLVE: _ClassVar[PagerDutyEventAction]

PAGER_DUTY_SEVERITY_UNSPECIFIED: PagerDutySeverity
PAGER_DUTY_SEVERITY_INFO: PagerDutySeverity
PAGER_DUTY_SEVERITY_WARNING: PagerDutySeverity
PAGER_DUTY_SEVERITY_ERROR: PagerDutySeverity
PAGER_DUTY_SEVERITY_CRITICAL: PagerDutySeverity
PAGER_DUTY_EVENT_ACTION_UNSPECIFIED: PagerDutyEventAction
PAGER_DUTY_EVENT_ACTION_TRIGGER: PagerDutyEventAction
PAGER_DUTY_EVENT_ACTION_ACKNOWLEDGE: PagerDutyEventAction
PAGER_DUTY_EVENT_ACTION_RESOLVE: PagerDutyEventAction

class PagerDutyEventV2Payload(_message.Message):
    __slots__ = ("summary", "timestamp", "severity", "source", "component", "group")
    SUMMARY_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    SEVERITY_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FIELD_NUMBER: _ClassVar[int]
    COMPONENT_FIELD_NUMBER: _ClassVar[int]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    CLASS_FIELD_NUMBER: _ClassVar[int]
    summary: str
    timestamp: _timestamp_pb2.Timestamp
    severity: PagerDutySeverity
    source: str
    component: str
    group: str
    def __init__(
        self,
        summary: _Optional[str] = ...,
        timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        severity: _Optional[_Union[PagerDutySeverity, str]] = ...,
        source: _Optional[str] = ...,
        component: _Optional[str] = ...,
        group: _Optional[str] = ...,
        **kwargs,
    ) -> None: ...

class PagerDutyEventV2Link(_message.Message):
    __slots__ = ("href", "text")
    HREF_FIELD_NUMBER: _ClassVar[int]
    TEXT_FIELD_NUMBER: _ClassVar[int]
    href: str
    text: str
    def __init__(self, href: _Optional[str] = ..., text: _Optional[str] = ...) -> None: ...

class PagerDutyEventV2Image(_message.Message):
    __slots__ = ("src", "href", "alt")
    SRC_FIELD_NUMBER: _ClassVar[int]
    HREF_FIELD_NUMBER: _ClassVar[int]
    ALT_FIELD_NUMBER: _ClassVar[int]
    src: str
    href: str
    alt: str
    def __init__(self, src: _Optional[str] = ..., href: _Optional[str] = ..., alt: _Optional[str] = ...) -> None: ...

class PagerDutyEventV2(_message.Message):
    __slots__ = ("payload", "routing_key", "event_action", "dedup_key", "client", "client_url", "links", "images")
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    ROUTING_KEY_FIELD_NUMBER: _ClassVar[int]
    EVENT_ACTION_FIELD_NUMBER: _ClassVar[int]
    DEDUP_KEY_FIELD_NUMBER: _ClassVar[int]
    CLIENT_FIELD_NUMBER: _ClassVar[int]
    CLIENT_URL_FIELD_NUMBER: _ClassVar[int]
    LINKS_FIELD_NUMBER: _ClassVar[int]
    IMAGES_FIELD_NUMBER: _ClassVar[int]
    payload: PagerDutyEventV2Payload
    routing_key: str
    event_action: PagerDutyEventAction
    dedup_key: str
    client: str
    client_url: str
    links: _containers.RepeatedCompositeFieldContainer[PagerDutyEventV2Link]
    images: _containers.RepeatedCompositeFieldContainer[PagerDutyEventV2Image]
    def __init__(
        self,
        payload: _Optional[_Union[PagerDutyEventV2Payload, _Mapping]] = ...,
        routing_key: _Optional[str] = ...,
        event_action: _Optional[_Union[PagerDutyEventAction, str]] = ...,
        dedup_key: _Optional[str] = ...,
        client: _Optional[str] = ...,
        client_url: _Optional[str] = ...,
        links: _Optional[_Iterable[_Union[PagerDutyEventV2Link, _Mapping]]] = ...,
        images: _Optional[_Iterable[_Union[PagerDutyEventV2Image, _Mapping]]] = ...,
    ) -> None: ...

class PagerDutyIntegration(_message.Message):
    __slots__ = ("id", "name", "default", "token", "environment_id")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DEFAULT_FIELD_NUMBER: _ClassVar[int]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    default: bool
    token: str
    environment_id: str
    def __init__(
        self,
        id: _Optional[str] = ...,
        name: _Optional[str] = ...,
        default: bool = ...,
        token: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
    ) -> None: ...

class IncidentIoIntegration(_message.Message):
    __slots__ = ("id", "token", "environment_id", "name", "source_id", "source_token", "severity_id", "severity_token")
    ID_FIELD_NUMBER: _ClassVar[int]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    SOURCE_ID_FIELD_NUMBER: _ClassVar[int]
    SOURCE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    SEVERITY_ID_FIELD_NUMBER: _ClassVar[int]
    SEVERITY_TOKEN_FIELD_NUMBER: _ClassVar[int]
    id: str
    token: str
    environment_id: str
    name: str
    source_id: str
    source_token: str
    severity_id: str
    severity_token: str
    def __init__(
        self,
        id: _Optional[str] = ...,
        token: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
        name: _Optional[str] = ...,
        source_id: _Optional[str] = ...,
        source_token: _Optional[str] = ...,
        severity_id: _Optional[str] = ...,
        severity_token: _Optional[str] = ...,
    ) -> None: ...

class IncidentIoEventV2(_message.Message):
    __slots__ = ("route_id", "route_token", "dedup_key", "source_url", "description", "status", "title")
    ROUTE_ID_FIELD_NUMBER: _ClassVar[int]
    ROUTE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    DEDUP_KEY_FIELD_NUMBER: _ClassVar[int]
    SOURCE_URL_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    TITLE_FIELD_NUMBER: _ClassVar[int]
    route_id: str
    route_token: str
    dedup_key: str
    source_url: str
    description: str
    status: str
    title: str
    def __init__(
        self,
        route_id: _Optional[str] = ...,
        route_token: _Optional[str] = ...,
        dedup_key: _Optional[str] = ...,
        source_url: _Optional[str] = ...,
        description: _Optional[str] = ...,
        status: _Optional[str] = ...,
        title: _Optional[str] = ...,
    ) -> None: ...

class TestPagerDutyIntegrationRequest(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class GetPagerDutyIntegrationRequest(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class GetPagerDutyIntegrationResponse(_message.Message):
    __slots__ = ("integration",)
    INTEGRATION_FIELD_NUMBER: _ClassVar[int]
    integration: PagerDutyIntegration
    def __init__(self, integration: _Optional[_Union[PagerDutyIntegration, _Mapping]] = ...) -> None: ...

class TestPagerDutyIntegrationResponse(_message.Message):
    __slots__ = ("status", "message")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    status: str
    message: str
    def __init__(self, status: _Optional[str] = ..., message: _Optional[str] = ...) -> None: ...

class AddPagerDutyIntegrationRequest(_message.Message):
    __slots__ = ("name", "token")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    name: str
    token: str
    def __init__(self, name: _Optional[str] = ..., token: _Optional[str] = ...) -> None: ...

class AddPagerDutyIntegrationResponse(_message.Message):
    __slots__ = ("integration",)
    INTEGRATION_FIELD_NUMBER: _ClassVar[int]
    integration: PagerDutyIntegration
    def __init__(self, integration: _Optional[_Union[PagerDutyIntegration, _Mapping]] = ...) -> None: ...

class DeletePagerDutyIntegrationRequest(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class DeletePagerDutyIntegrationResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class UpdatePagerDutyIntegrationRequest(_message.Message):
    __slots__ = ("id", "name", "default", "token")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DEFAULT_FIELD_NUMBER: _ClassVar[int]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    default: bool
    token: str
    def __init__(
        self, id: _Optional[str] = ..., name: _Optional[str] = ..., default: bool = ..., token: _Optional[str] = ...
    ) -> None: ...

class UpdatePagerDutyIntegrationResponse(_message.Message):
    __slots__ = ("integration",)
    INTEGRATION_FIELD_NUMBER: _ClassVar[int]
    integration: PagerDutyIntegration
    def __init__(self, integration: _Optional[_Union[PagerDutyIntegration, _Mapping]] = ...) -> None: ...

class SetDefaultPagerDutyIntegrationRequest(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class SetDefaultPagerDutyIntegrationResponse(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class GetAllPagerDutyIntegrationsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetAllPagerDutyIntegrationsResponse(_message.Message):
    __slots__ = ("integrations",)
    INTEGRATIONS_FIELD_NUMBER: _ClassVar[int]
    integrations: _containers.RepeatedCompositeFieldContainer[PagerDutyIntegration]
    def __init__(self, integrations: _Optional[_Iterable[_Union[PagerDutyIntegration, _Mapping]]] = ...) -> None: ...

class TestIncidentIoIntegrationRequest(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class TestIncidentIoIntegrationResponse(_message.Message):
    __slots__ = ("status", "message")
    STATUS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    status: str
    message: str
    def __init__(self, status: _Optional[str] = ..., message: _Optional[str] = ...) -> None: ...

class GetIncidentIoIntegrationRequest(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class GetIncidentIoIntegrationResponse(_message.Message):
    __slots__ = ("integration",)
    INTEGRATION_FIELD_NUMBER: _ClassVar[int]
    integration: IncidentIoIntegration
    def __init__(self, integration: _Optional[_Union[IncidentIoIntegration, _Mapping]] = ...) -> None: ...

class AddIncidentIoIntegrationRequest(_message.Message):
    __slots__ = ("integration_name", "integration_token")
    INTEGRATION_NAME_FIELD_NUMBER: _ClassVar[int]
    INTEGRATION_TOKEN_FIELD_NUMBER: _ClassVar[int]
    integration_name: str
    integration_token: str
    def __init__(self, integration_name: _Optional[str] = ..., integration_token: _Optional[str] = ...) -> None: ...

class AddIncidentIoIntegrationResponse(_message.Message):
    __slots__ = ("integration",)
    INTEGRATION_FIELD_NUMBER: _ClassVar[int]
    integration: IncidentIoIntegration
    def __init__(self, integration: _Optional[_Union[IncidentIoIntegration, _Mapping]] = ...) -> None: ...

class DeleteIncidentIoIntegrationRequest(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class DeleteIncidentIoIntegrationResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class UpdateIncidentIoIntegrationRequest(_message.Message):
    __slots__ = ("id", "name", "token")
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    id: str
    name: str
    token: str
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., token: _Optional[str] = ...) -> None: ...

class UpdateIncidentIoIntegrationResponse(_message.Message):
    __slots__ = ("integration",)
    INTEGRATION_FIELD_NUMBER: _ClassVar[int]
    integration: IncidentIoIntegration
    def __init__(self, integration: _Optional[_Union[IncidentIoIntegration, _Mapping]] = ...) -> None: ...

class GetAllIncidentIoIntegrationsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetAllIncidentIoIntegrationsResponse(_message.Message):
    __slots__ = ("integrations",)
    INTEGRATIONS_FIELD_NUMBER: _ClassVar[int]
    integrations: _containers.RepeatedCompositeFieldContainer[IncidentIoIntegration]
    def __init__(self, integrations: _Optional[_Iterable[_Union[IncidentIoIntegration, _Mapping]]] = ...) -> None: ...



import json
import os
from typing import Any, Callable, Mapping, Optional, Tuple, TypeVar, overload

T = TypeVar("T")


def has_integration(integration_name: str) -> bool:
    encoded = os.getenv("_CHALK_AVAILABLE_INTEGRATIONS")
    if encoded is not None:
        available = set(json.loads(encoded))
        return integration_name in available
    return False


def _get_integration_variable_name(name: str, integration_name: Optional[str]) -> str:
    return name if integration_name is None else f"{integration_name}_{name}"


def _serialize_integration_value_to_str(val: Any) -> str:
    if isinstance(val, bytes):
        return val.decode("utf-8")
    return str(val)


@overload
def load_integration_variable(
    name: str,
    integration_name: Optional[str],
    override: Optional[Mapping[str, str]],
) -> Optional[str]:
    ...


@overload
def load_integration_variable(
    name: str, integration_name: Optional[str], override: Optional[Mapping[str, str]], parser: Callable[[str], T]
) -> Optional[T]:
    ...


def load_integration_variable(
    name: str,
    integration_name: Optional[str],
    override: Optional[Mapping[str, str]],
    parser: Callable[[str], Any] = str,
) -> Optional[Any]:
    integration_var_name = _get_integration_variable_name(name, integration_name)
    if override is not None:
        value = override.get(integration_var_name, os.getenv(integration_var_name))
    else:
        value = os.getenv(integration_var_name)

    return parser(value) if value is not None else None


def create_integration_variable(
    name: str,
    integration_name: Optional[str],
    value: Any,
    serializer: Callable[[Any], str] = _serialize_integration_value_to_str,
) -> Tuple[str, Optional[str]]:
    return _get_integration_variable_name(name, integration_name), serializer(value) if value is not None else None
